# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Attack related factories are defined here
"""
import random

from pyherc.aspects import log_debug, log_info
from pyherc.rules.factory import SubActionFactory
from pyherc.rules.moving.action import (EscapeAction, MoveAction,
                                        SwitchPlacesAction)


class WalkFactory(SubActionFactory):
    """
    Factory for creating walk actions
    """
    @log_debug
    def __init__(self, level_generator_factory, dying_rules):
        """
        Constructor for this factory

        :param level_generator_factory: Factory for generating level generators
        :type level_generator_factory: LevelGeneratorFactory
        """
        super().__init__()
        self.level_generator_factory = level_generator_factory
        self.dying_rules = dying_rules
        self.movement_mode = 'walk'

    @log_debug
    def __str__(self):
        return 'walk factory'

    @log_debug
    def can_handle(self, parameters):
        """
        Can this factory process these parameters

        :param parameters: Parameters to check
        :type parameters: MoveParameters

        :returns: True if factory is capable of handling parameters
        :rtype: boolean
        """
        return self.movement_mode == parameters.movement_mode

    @log_info
    def get_action(self, parameters):
        """
        Create a walk action

        :param parameters: Parameters used to control walk creation
        :type parameters: MoveParameters
        """
        location = parameters.character.location
        new_level = parameters.character.level
        direction = parameters.direction
        new_location = None

        if direction == 1:
            new_location = (location[0], location[1] - 1)
        elif direction == 2:
            new_location = (location[0] + 1, location[1] - 1)
        elif direction == 3:
            new_location = (location[0] + 1, location[1])
        elif direction == 4:
            new_location = (location[0] + 1, location[1] + 1)
        elif direction == 5:
            new_location = (location[0], location[1] + 1)
        elif direction == 6:
            new_location = (location[0] - 1, location[1] + 1)
        elif direction == 7:
            new_location = (location[0] - 1, location[1])
        elif direction == 8:
            new_location = (location[0] - 1, location[1] - 1)
        elif direction == 9:
            return self._get_action_for_portal(parameters.character)
        else:
            raise RuntimeError('Character does not know where to go')

        if new_level.blocks_movement(new_location[0], new_location[1]):
            new_location = parameters.character.location
        elif new_level.get_creature_at(new_location):
            return self.get_place_switch_action(parameters.character,
                                                new_location)

        return MoveAction(character=parameters.character,
                          new_location=new_location,
                          new_level=new_level,
                          skip_creature_check=False,
                          dying_rules=self.dying_rules)

    @log_debug
    def get_place_switch_action(self, character, new_location):
        """
        Get action for two character switching places

        .. versionadded:: 0.11
        """
        other_character = character.level.get_creature_at(new_location)

        return SwitchPlacesAction(character,
                                  other_character,
                                  self.dying_rules)

    @log_debug
    def _get_action_for_portal(self, character):
        """
        Get location after entering a portal

        :returns: action for entering portal
        :rtype: Action

        .. versionadded:: 0.11

        .. note:: Entering portal may lead player next to other end if it is
                  blocked by other character
        """
        location = character.location
        new_level = character.level
        new_location = None

        portal = new_level.get_portal_at(location)
        if portal is not None:
            if not portal.exits_dungeon:
                other_end = portal.get_other_end(self.level_generator_factory)
                if other_end is not None:
                    new_level = other_end.level
                    new_location = self._area_around_portal(other_end)
                else:
                    raise RuntimeError('Portal leads to void!')
            else:
                return EscapeAction(character)
        else:
            new_level = character.level
            new_location = character.location

        return MoveAction(character,
                          new_location,
                          new_level,
                          False,
                          self.dying_rules)

    @log_debug
    def _area_around_portal(self, portal):
        """
        Get passable tiles in 3x3 area around a portal
        """
        level = portal.level
        location = portal.location

        if not (level.blocks_movement(location[0], location[1]) or
                level.get_creature_at(location)):
            return location

        passables = []

        for loc_x in range(location[0]-1, location[1]+1):
            for loc_y in range(location[1]-1, location[1]+1):
                if not (level.blocks_movement(loc_x, loc_y) or
                        level.get_creature_at((loc_x, loc_y))):
                    passables.append((loc_x, loc_y))

        return random.choice(passables)


class MoveFactory(SubActionFactory):
    """
    Factory for constructing move actions
    """
    @log_debug
    def __init__(self, factories):
        """
        Constructor for this factory

        :param factories: a single Factory or list of Factories to use
        """
        super().__init__()
        self.action_type = 'move'

        if hasattr(factories, '__iter__'):
            self.factories = factories
        else:
            self.factories = []
            self.factories.append(factories)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Factory class for animations
"""

from .animation import Animation
from .attack import AttackHitAnimation
from .damage import DamageTriggeredAnimation
from .death import DeathAnimation
from .healing import HealAddedAnimation, HealTriggeredAnimation
from .inventory import DropAnimation, PickUpAnimation
from .moving import MoveAnimation
from .perception import NoticeAnimation, LoseFocusAnimation
from .poison import PoisonAddedAnimation, PoisonTriggeredAnimation


class AnimationFactory():
    """
    Class for creating animations for events

    .. versionadded:: 0.12
    """
    def __init__(self):
        """
        Default constructor
        """
        self.animations = {
            'attack hit': AttackHitAnimation,
            'damage triggered': DamageTriggeredAnimation,
            'death': DeathAnimation,
            'drop': DropAnimation,
            'heal started': HealAddedAnimation,
            'heal triggered': HealTriggeredAnimation,
            'lose focus': LoseFocusAnimation,
            'move': MoveAnimation,
            'notice': NoticeAnimation,
            'pick up': PickUpAnimation,
            'poisoned': PoisonAddedAnimation,
            'poison triggered': PoisonTriggeredAnimation
        }

    def create_animation(self, event):
        """
        Create an animation for event

        :param event: event to create animation for
        :type event: Event
        :returns: animation
        :rtype: Animation
        """
        if event.event_type in self.animations:
            return self.animations[event.event_type](event)
        else:
            return Animation(event)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Public interface for combat rules
"""

from pyherc.aspects import log_debug, log_info
from pyherc.rules.public import ActionParameters


@log_info
def attack(character, direction, action_factory, rng):
    """
    Attack to given direction

    :param direction: direction to attack
    :type direction: integer
    :param action_factory: factory to create actions
    :type action_factory: ActionFactory
    :param rng: random number generator
    :type rng: Random
    """
    if character.inventory.weapon is not None:
        weapon = character.inventory.weapon.weapon_data
        if character.inventory.projectiles is not None:
            ammunition = character.inventory.projectiles.ammunition_data
        else:
            ammunition = None
    else:
        weapon = None

    if weapon is None:
        attack_type = 'unarmed'
    else:
        if (ammunition is None or
                weapon.ammunition_type != ammunition.ammunition_type):
            attack_type = 'melee'
        else:
            target_loc = character.get_location_at_direction(direction)
            if character.level.get_creature_at(target_loc) is None:
                if character.level.blocks_movement(target_loc[0], target_loc[1]):
                    attack_type = 'melee'
                else:
                    attack_type = 'ranged'
            else:
                attack_type = 'melee'

    action = action_factory.get_action(
        AttackParameters(attacker=character,
                         direction=direction,
                         attack_type=attack_type,
                         random_number_generator=rng))

    action.execute()


class AttackParameters(ActionParameters):
    """
    Object for controlling attack action creation
    """

    @log_debug
    def __init__(self, attacker, direction, attack_type,
                 random_number_generator):
        """
        Construct AttackParameters

        Args:
            attacker: Character doing an attack
            direction: Direction to attack to
            attack_type: type of attack to perform
            random_number_generator: Random number generator to use
        """
        super().__init__()

        self.action_type = 'attack'
        self.attacker = attacker
        self.direction = direction
        self.attack_type = attack_type
        self.random_number_generator = random_number_generator
        self.model = None

    @log_debug
    def __str__(self):
        """
        Get string representation of this object
        """
        return 'attack with attack type of ' + self.attack_type



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module for handling loading of images and icons
"""
import curses

from pyherc.aspects import log_debug


class CursesSurfaceManager():
    """
    Class for managing glyphs

    .. versionadded:: 0.9
    """
    @log_debug
    def __init__(self):
        """
        Default constructor
        """
        super().__init__()
        self.resourcesLoaded = 0
        self.icons = {}
        self.attributes = {}

    @log_debug
    def load_resources(self):
        """
        Load graphics from files
        """
        pass

    @log_debug
    def add_icon(self, key, filename, ascii_char, attributes = None):
        """
        Add icon to internal collection
        """
        if not (hasattr(key, 'upper') or hasattr(key, 'real')):
            key = key[0]

        if not key in self.icons:
            self.icons[key] = ascii_char

            if attributes == None:
                used_attributes = curses.A_NORMAL | self.get_attribute_by_name('white')
            else:
                used_attributes = curses.A_NORMAL
                for elem in attributes:
                    used_attributes = used_attributes | self.get_attribute_by_name(elem)

            self.attributes[key] = used_attributes

        return key

    def get_attribute_by_name(self, attribute_name):
        """
        Get attribute based on name
        """
        if attribute_name == 'blue':
            return curses.color_pair(1)
        elif attribute_name == 'cyan':
            return curses.color_pair(2)
        elif attribute_name == 'green':
            return curses.color_pair(3)
        elif attribute_name == 'magenta':
            return curses.color_pair(4)
        elif attribute_name == 'red':
            return curses.color_pair(5)
        elif attribute_name == 'white':
            return curses.color_pair(6)
        elif attribute_name == 'yellow':
            return curses.color_pair(7)
        elif attribute_name == 'bold':
            return curses.A_BOLD
        elif attribute_name == 'normal':
            return curses.A_NORMAL
        elif attribute_name == 'dim':
            return curses.A_DIM

    def get_attribute(self, id):
        """
        Get attriutes with ID
        """
        if id in self.attributes:
            return self.attributes[id]
        else:
            return curses.A_NORMAL | curses.color_pair(6)

    def get_icon(self, id):
        """
        Get icon with ID

        :param id: ID number of the icon to retrieve
        :type id: int
        :returns: icon if found, otherwise empty icon
        :rtype: string
        """
        if not id:
            return ' '

        if not (hasattr(id, 'upper') or hasattr(id, 'real')):
            id = id[0]

        if id in self.icons:
            return self.icons[id]
        else:
            return 'x'



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module defining classes related to Move
"""
from pyherc.aspects import log_debug, log_info
from pyherc.data.constants import Duration
from pyherc.data.geometry import find_direction
from pyherc.data.model import ESCAPED_DUNGEON
from pyherc.events import MoveEvent


class MoveAction():
    """
    Action for moving
    """
    @log_debug
    def __init__(self, character, new_location, new_level,
                 skip_creature_check, dying_rules):
        """
        Default constructor

        :param character: character moving
        :type character: Character
        :param new_location: location to move
        :type new_location: (int, int)
        :param new_level: level to move
        :type new_level: Level
        :param skip_creature_check: bypass checking if other creature blocks
        :type skip_creature_check: boolean
        :param dying_rules: rules for dying
        """
        super().__init__()

        self.character = character
        self.new_location = new_location
        self.new_level = new_level
        self.model = None
        self.skip_creature_check = skip_creature_check
        self.dying_rules = dying_rules

    @log_info
    def execute(self):
        """
        Executes this Move
        """
        if self.is_legal():

            affected_tiles = [self.character.location,
                              self.new_location]

            old_location = self.character.location
            self.character.location = self.new_location
            direction = find_direction(old_location, self.new_location)

            if self.new_level is not None:
                if self.character.level != self.new_level:
                    self.character.level.remove_creature(self.character)
                    self.character.level = self.new_level
                    self.new_level.add_creature(self.character,
                                                self.new_location)

            armour = self.character.inventory.armour
            if armour:
                speed_modifier = armour.armour_data.speed_modifier
            else:
                speed_modifier = 1

            self.character.add_to_tick(Duration.fast / speed_modifier)

            self.character.raise_event(MoveEvent(
                mover=self.character,
                old_location=old_location,
                direction=direction,
                affected_tiles=affected_tiles))

        else:
            self.character.add_to_tick(Duration.instant)

        traps = self.character.level.get_traps(self.character.location)

        for trap in traps:
            trap.on_enter(self.character)

        self.dying_rules.check_dying(self.character)

    @log_debug
    def is_legal(self):
        """
        Check if the move is possible to perform

        :returns: True if move is possible, false otherwise
        :rtype: Boolean
        """
        location_ok = True
        if self.new_level is not None:
            if not self.new_level.blocks_movement(self.new_location[0],
                                                  self.new_location[1]):
                if (not self.skip_creature_check and
                        self.new_level.get_creature_at(self.new_location)):
                    location_ok = False
            else:
                location_ok = False
        else:
            location_ok = False

        return location_ok

    def __str__(self):
        """
        String representation of this action
        """
        return '{0} at {1}:{2}'.format(self.character,
                                       self.new_location,
                                       self.new_level)


class WalkAction(MoveAction):
    """
    Action for walking
    """
    @log_info
    def execute(self):
        """
        Execute this move
        """
        super().execute()


class EscapeAction(MoveAction):
    """
    Action for escaping the dungeon

    .. versionadded:: 0.8
    """
    @log_debug
    def __init__(self, character):
        """
        Default constructor
        """
        super().__init__(character=character,
                         new_location=None,
                         new_level=None,
                         skip_creature_check=False,
                         dying_rules=None)

    @log_info
    def execute(self):
        """
        Execute this move
        """
        model = self.character.model
        model.end_condition = ESCAPED_DUNGEON

    @log_debug
    def is_legal(self):
        """
        Check if the move is possible to perform

        :returns: True if move is possible, false otherwise
        :rtype: Boolean
        """
        model = self.character.model

        if model.player == self.character:
            return True
        else:
            return False


class SwitchPlacesAction():
    """
    Action for switching places with another creature

    .. versionadded:: 0.11
    """

    @log_debug
    def __init__(self, character, other_character, dying_rules):
        """
        Default constructor

        :param character: character to move
        :type character: Character
        :param other_character: other character to move
        :type other_character: Character
        """
        assert character is not None
        assert other_character is not None

        self.character = character
        self.other_character = other_character

        self.move_action_1 = WalkAction(self.character,
                                        self.other_character.location,
                                        self.other_character.level,
                                        True,
                                        dying_rules)

        self.move_action_2 = WalkAction(self.other_character,
                                        self.character.location,
                                        self.character.level,
                                        True,
                                        dying_rules)

    @log_info
    def execute(self):
        """
        Execute this move
        """
        if self.is_legal():
            self.move_action_1.execute()
            self.move_action_2.execute()
        else:
            self.character.add_to_tick(Duration.instant)

    @log_debug
    def is_legal(self):
        """
        Check if the move is possible to perform

        :returns: True if move is possible, false otherwise
        :rtype: Boolean
        """
        model = self.character.model

        if model.player in (self.character, self.other_character):
            return False
        else:
            return (self.move_action_1.is_legal() and
                    self.move_action_2.is_legal())



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Utility classes

BSPSection
"""

import collections
import logging
import random


class BSPSection():
    """
    Class used to divide area in sections
    """
    def __init__(self, corner1=None, corner2=None, parent=None,
                 direction=None):
        """
        Default constructor
        """
        self.corner1 = corner1
        self.corner2 = corner2
        self.node1 = None
        self.node2 = None
        self.parent = parent
        self.direction = direction
        self.logger = logging.getLogger('pyherc.generators.utils.BSPSection')

    def split(self, min_size=(6, 6), direction=None):
        """
        Split BSPSection in two
        Links two new BSPSections into this one
        :param min_size: minimum size to split into
        :param direction: horizontal (1) / vertical split (2)
        """
        assert self.corner1 is not None
        assert self.corner2 is not None

        size = (abs(self.corner1[0] - self.corner2[0]),
                abs(self.corner1[1] - self.corner2[1]))

        if direction is None:
            # 1 split horizontal, 2 split vertical
            direction = random.randint(1, 2)
        assert(direction in (1, 2))

        if direction == 1:
            if size[1] < 2 * min_size[1]:
                if size[0] < 2 * min_size[0]:
                    return None
                else:
                    direction = 2
        else:
            if size[0] < 2 * min_size[0]:
                if size[1] < 2 * min_size[1]:
                    return None
                else:
                    direction = 1

        if direction == 1:
            split_location = random.randint(min_size[1], size[1] - min_size[1])
            self.node1 = BSPSection(self.corner1,
                                    (self.corner2[0],
                                        self.corner1[1] + split_location),
                                    self, direction)
            self.node2 = BSPSection((self.corner1[0],
                                     self.corner1[1] + split_location + 1),
                                    self.corner2,
                                    self, direction)
        else:
            split_location = random.randint(min_size[0], size[0] - min_size[0])
            self.node1 = BSPSection(self.corner1,
                                    (self.corner1[0] + split_location,
                                     self.corner2[1]),
                                    self, direction)
            self.node2 = BSPSection((self.corner1[0] + split_location + 1,
                                    self.corner1[1]),
                                    self.corner2,
                                    self, direction)

    def __str__(self):
        """
        Override __str__ to print meaningful string representation
        """
        if self.corner1 is not None:
            if self.corner2 is not None:
                return self.corner1.__str__() + ':' + self.corner2.__str__()
            else:
                return self.corner1.__str__() + ':None'
        else:
            if self.corner2 is not None:
                return 'None:' + self.corner2.__str__()
            else:
                return 'None:None'

    def get_area_queue(self):
        """
        Gets list of BSPSections, starting from leaves and moving towards root
        """
        queue = collections.deque()
        list = []
        queue.append(self)

        while len(queue) > 0:
            item = queue.pop()
            list.append(item)
            if item.node1 is not None:
                queue.appendleft(item.node1)
            if item.node2 is not None:
                queue.appendleft(item.node2)

        list.reverse()

        return list

    def get_center(self):
        """
        Calculates center of the BSPSection
        :returns: center point
        """
        assert self.corner1 is not None
        assert self.corner2 is not None
        center = (int(self.corner1[0] +
                  ((self.corner2[0] - self.corner1[0]) / 2)),
                  int(self.corner1[1] +
                      ((self.corner2[1] - self.corner1[1]) / 2)))

        return center



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Events for hit points
"""
from pyherc.events.event import Event


class HitPointsChangedEvent(Event):
    """
    Event to raise when hit points change

    .. versionadded:: 0.7
    """
    def __init__(self, character, old_hit_points, new_hit_points):
        """
        Default constructor
        """
        super().__init__(event_type='hit points changed',
                         level=character.level,
                         location=character.location,
                         affected_tiles=[])

        self.character = character
        self.old_hit_points = old_hit_points
        self.new_hit_points = new_hit_points



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Public interface of move package

.. versionadded:: 0.10
"""

from pyherc.aspects import log_debug, log_info
from pyherc.rules.public import ActionParameters


@log_info
def move(character, direction, action_factory):
    """
    Move character to specified direction

    :param character: character moving
    :type character: Character
    :param direction: direction to move
    :type direction: integer
    :param action_factory: factory to create actions
    :type action_factory: ActionFactory
    """
    action = action_factory.get_action(MoveParameters(character,
                                                      direction,
                                                      'walk'))
    action.execute()


@log_debug
def is_move_legal(character, direction, movement_mode, action_factory):
    """
    Check if movement is legal

    :param character: character about to move
    :type character: Character
    :param direction: direction to move
    :type direction: integer
    :param movement_mode: mode of movement
    :type movement_mode: string
    :param action_factory: factory to create actions
    :type action_factory: ActionFactory
    :returns: True if move is legal, False otherwise
    :rtype: Boolean
    """
    action = action_factory.get_action(MoveParameters(character,
                                                      direction,
                                                      movement_mode))
    return action.is_legal()


class MoveParameters(ActionParameters):
    """
    Object for controlling move action creation
    """
    @log_debug
    def __init__(self, character, direction, movement_mode):
        """
        Construct move parameters

        Args:
            character: Character moving
            direction: Direction of the move
            movement_mode: Mode of movement
        """
        super().__init__()

        self.action_type = 'move'
        self.character = character
        self.direction = direction
        self.movement_mode = movement_mode
        self.model = None

    @log_debug
    def __str__(self):
        """
        Get string representation of this object
        """
        return 'move with movement mode of ' + self.movement_mode



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes for inventory animations
"""

from .animation import Animation
from herculeum.ui.gui.layers import zorder_item


class DropAnimation(Animation):
    """
    Generic animation for dropping item

    .. versionadded:: 0.12
    """
    def __init__(self, event):
        """
        Default constructor
        """
        super().__init__(event)

        self.item = event.item

    def trigger(self, ui):
        """
        Trigger this animation
        """
        ui.add_glyph(self.item, ui.scene, zorder_item)


class PickUpAnimation(Animation):
    """
    Generic animation for picking up an item

    .. versionadded:: 0.12
    """
    def __init__(self, event):
        """
        Default constructor
        """
        super().__init__(event)

        self.item = event.item

    def trigger(self, ui):
        """
        Trigger this animation
        """
        glyphs = [x for x in ui.view.items()
                  if (hasattr(x, 'entity'))
                  and (x.entity == self.item)]

        for glyph in glyphs:
            ui.view.scene().removeItem(glyph)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Public interface for consuming actions
"""

from pyherc.aspects import log_debug, log_info
from pyherc.rules.public import ActionParameters


@log_info
def drink(character, potion, action_factory):
    """
    Drink potion

    :param character: character about to drink
    :type character: Character
    :param potion: potion to drink
    :type potion: Item
    :param action_factory: factory to create actions
    :type action_factory: ActionFactory
    """
    action = action_factory.get_action(DrinkParameters(character,
                                                       potion))
    action.execute()


class DrinkParameters(ActionParameters):
    """
    Object for controlling drink action creation
    """
    @log_debug
    def __init__(self, character, item):
        """
        Construct drink parameters

        Args:
            character: Character moving
            item: Item to drink
        """
        super().__init__()

        self.action_type = 'drink'
        self.character = character
        self.item = item
        self.model = None



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes needed for item generation
"""

import random

from pyherc.aspects import log_debug, log_info
from pyherc.data import Item
from pyherc.data.effects import EffectHandle, EffectsCollection
from pyherc.data.item import AmmunitionData, ArmourData, WeaponData


class ItemGenerator():
    """
    Class used to generate items
    """
    @log_debug
    def __init__(self, config):
        """
        Default constructor

        :param config: configuration for items
        :type config: ItemConfiguration
        """
        super().__init__()
        self.config = config

    @log_info
    def generate_item(self, name=None, item_type=None):
        """
        Generate an item

        :param name: name of the item to generate
        :type name: string
        :param item_type: type of the item to generate
        :type item_type: string
        :return: Generated item
        :rtype: Item
        """
        item_specification = self.find_item_specification(name=name,
                                                          item_type=item_type)

        item = self.create_item(item_specification)

        return item

    @log_debug
    def find_item_specification(self, name=None, item_type=None):
        """
        Find item specification by given parameters

        :param name: name of the item
        :type name: string
        :param item_type: type of the item to generate
        :type item_type: string
        :return: item specification
        :rtype: ItemConfiguration
        """
        if name is not None:
            return self.config.get_by_name(name)
        else:
            return self.config.get_by_type(item_type)

    @log_debug
    def create_item(self, item_specification):
        """
        Create new item based on specification

        :param item_specification: specification of item
        :type item_specification: ItemConfiguration
        :return: new item
        :rtype: Item
        """
        item = Item(EffectsCollection())

        item.name = item_specification.name
        item.description = item_specification.description
        item.icon = random.choice(item_specification.icons)
        item.cost = item_specification.cost
        item.weight = item_specification.weight
        item.rarity = item_specification.rarity
        item.tags = item_specification.types

        if not item_specification.weapon_configration is None:
            weapon_spec = item_specification.weapon_configration
            damage = []
            damage.extend(weapon_spec.damage)
            item.weapon_data = WeaponData(
                damage=(damage),
                critical_range=weapon_spec.critical_range,
                critical_damage=weapon_spec.critical_damage,
                weapon_type=weapon_spec.weapon_class,
                ammunition_type=weapon_spec.ammunition_type,
                speed=weapon_spec.speed)

        if item_specification.armour_configuration is not None:
            armour_spec = item_specification.armour_configuration
            item.armour_data = ArmourData(
                damage_reduction=armour_spec.damage_reduction,
                speed_modifier=armour_spec.speed_modifier)

        if item_specification.ammunition_configuration is not None:
            ammo_spec = item_specification.ammunition_configuration
            item.ammunition_data = AmmunitionData(
                damage=ammo_spec.damage,
                critical_range=ammo_spec.critical_range,
                critical_damage=ammo_spec.critical_damage,
                ammunition_type=ammo_spec.ammunition_type,
                count=ammo_spec.count)

        for spec in item_specification.effect_handles:
            new_handle = EffectHandle(trigger=spec.trigger,
                                      effect=spec.effect,
                                      parameters=spec.parameters,
                                      charges=spec.charges)

            item.add_effect_handle(new_handle)

        return item


class ItemConfigurations():
    """
    Class for configuring items
    """
    @log_debug
    def __init__(self, rng):
        """
        Default constructor

        :param rng: random number generator
        :type rng: Random
        """
        super().__init__()
        self.__items = []
        self.__items_by_name = {}
        self.rng = rng

    @log_debug
    def add_item(self, item_config):
        """
        Add item to internal configuration
        """
        self.__items.append(item_config)
        self.__items_by_name[item_config.name] = item_config

    @log_debug
    def get_all_items(self):
        """
        Returns all items

        :return: list of items
        :rtype: [ItemConfiguration]
        """
        return self.__items

    @log_debug
    def get_by_name(self, name):
        """
        Retrieve specification of item by name

        :param name: name of the item
        :type name: string
        :return: item specification
        :rtype: ItemConfiguration
        """
        return self.__items_by_name[name]

    @log_debug
    def get_by_type(self, item_type):
        """
        Retrieve a random specification of item by type

        :param item_type: type of the item
        :type item_type: string
        :return: item specification
        :rtype: ItemConfiguration
        """
        matching_specs = [x for x in self.__items
                          if item_type in x.types]

        max_score = sum([x.rarity for x in matching_specs])

        score = self.rng.randint(1, max_score)
        current_score = 0

        for spec in matching_specs:
            current_score = current_score + spec.rarity
            if current_score >= score:
                return spec

        return None


class ItemConfiguration():
    """
    Class representing a single item
    """
    rarities = {'artifact': 1,
                'legendary': 4,
                'epic': 16,
                'rare': 64,
                'uncommon': 256,
                'common': 1024}

    @log_debug
    def __init__(self, name, cost, weight, icons, types, rarity,
                 weapon_configration=None, effect_handles=None,
                 armour_configuration=None,
                 ammunition_configuration=None, description=''):
        """
        Default constructor
        """
        super().__init__()
        self.name = name
        self.description = description
        self.cost = cost
        self.weight = weight
        self.icons = icons
        self.types = types
        self.rarity = self.rarities[rarity]
        self.weapon_configration = weapon_configration
        self.armour_configuration = armour_configuration
        self.ammunition_configuration = ammunition_configuration

        if effect_handles is None:
            self.effect_handles = []
        else:
            self.effect_handles = effect_handles


class WeaponConfiguration():
    """
    Class representing weapon configuration
    """
    @log_debug
    def __init__(self, damage, critical_range, critical_damage, weapon_class,
                 ammunition_type=None, speed=1):
        """
        Default constructor
        """
        super().__init__()

        self.damage = damage
        self.critical_range = critical_range
        self.critical_damage = critical_damage
        self.weapon_class = weapon_class
        self.ammunition_type = ammunition_type
        self.speed = speed


class ArmourConfiguration():
    """
    Class representing armour configuration
    """
    @log_debug
    def __init__(self, damage_reduction, speed_modifier):
        """
        Default constructor
        """
        super().__init__()

        self.damage_reduction = damage_reduction
        self.speed_modifier = speed_modifier


class AmmunitionConfiguration():
    """
    Class representing ammunition configuration

    .. versionadded:: 0.8
    """
    @log_debug
    def __init__(self, damage, critical_range, critical_damage,
                 ammunition_type, count):
        """
        Default constructor
        """
        super().__init__()

        self.damage = damage
        self.critical_range = critical_range
        self.critical_damage = critical_damage
        self.ammunition_type = ammunition_type
        self.count = count



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
module for configuring first gate
"""
import hy
from herculeum.ai import FlockingHerbivore, SkeletonWarriorAI
from herculeum.ai.firebeetle import FireBeetleAI
from herculeum.ai.rat import RatAI
from pyherc.config.dsl import LevelConfiguration, LevelContext
from pyherc.data.effects import DamageModifier
from pyherc.generators import creature_config, inventory_config
from pyherc.generators.level.creatures import (CreatureAdder,
                                               CreatureAdderConfiguration)
from pyherc.generators.level.decorator import (AggregateDecorator,
                                               AggregateDecoratorConfig,
                                               DirectionalWallDecorator,
                                               DirectionalWallDecoratorConfig,
                                               FloorBuilderDecorator,
                                               FloorBuilderDecoratorConfig,
                                               ReplacingDecorator,
                                               ReplacingDecoratorConfig,
                                               WallBuilderDecorator,
                                               WallBuilderDecoratorConfig,
                                               WallOrnamentDecorator,
                                               WallOrnamentDecoratorConfig)
from pyherc.generators.level.items import ItemAdder, ItemAdderConfiguration
from pyherc.generators.level.partitioners import GridPartitioner
from pyherc.generators.level.portals import PortalAdderConfiguration
from pyherc.generators.level.room import (PillarRoomGenerator,
                                          SquareRoomGenerator)
from pyherc.rules.constants import (CRUSHING_DAMAGE, LIGHT_DAMAGE,
                                    PIERCING_DAMAGE, POISON_DAMAGE)


def init_level(rng, item_generator, creature_generator, level_size, context):
    """
    Initialise upper catacombs

    :returns: level configuration
    :rtype: LevelConfiguration
    """
    surface_manager = context.surface_manager

    tile_floor = 'tile floor'

    wall_empty = None
    floor_empty = None
    wall_natural = 'natural wall'
    wall_constructed = 'constructed wall'
    wall = 'wall'

    stairs_down = surface_manager.add_icon('stairs_down',
                                           ':stairs_down.png',
                                           '>')
    stairs_up = surface_manager.add_icon('stairs_up',
                                         ':stairs_up.png',
                                         '<')

    pillar = surface_manager.add_icon('pillar', ':pillar.png', '#')

    wall_15 = surface_manager.add_icon('catacombs_wall_15', ':catacombs_wall_2_8.png', '#')
    wall_57 = surface_manager.add_icon('catacombs_wall_57', ':catacombs_wall_2_4.png', '#')
    wall_35 = surface_manager.add_icon('catacombs_wall_35', ':catacombs_wall_2_6.png', '#')
    wall_37 = surface_manager.add_icon('catacombs_wall_37', ':catacombs_wall_4_6.png', '#')
    wall_17 = surface_manager.add_icon('catacombs_wall_17', ':catacombs_wall_8_4.png', '#')
    wall_13 = surface_manager.add_icon('catacombs_wall_13', ':catacombs_wall_8_6.png', '#')

    wall_135 = surface_manager.add_icon('catacombs_wall_135', ':catacombs_wall_2_6_8.png', '#')
    wall_357 = surface_manager.add_icon('catacombs_wall_357', ':catacombs_wall_2_4_6.png', '#')
    wall_1357 = surface_manager.add_icon('catacombs_wall_1357', ':catacombs_wall_2_4_6_8.png', '#')
    wall_137 = surface_manager.add_icon('catacombs_wall_137', ':catacombs_wall_4_6_8.png', '#')
    wall_157 = surface_manager.add_icon('catacombs_wall_157', ':catacombs_wall_2_4_8.png', '#')

    room_generators = [SquareRoomGenerator(tile_floor,
                                           wall_empty,
                                           tile_floor,
                                           ['first gate']),
                       PillarRoomGenerator(floor_tile = tile_floor,
                                           corridor_tile = tile_floor,
                                           empty_tile = wall_empty,
                                           pillar_tile = pillar,
                                           level_types = ['first gate'])]

    level_partitioners = [GridPartitioner(['first gate'],
                                           3,
                                           2,
                                           rng)]

    wallbuilder_config = WallBuilderDecoratorConfig(['first gate'],
                                        {wall_natural: wall_constructed},
                                         wall_empty)
    wallbuilder = WallBuilderDecorator(wallbuilder_config)

    wall_direction_config = DirectionalWallDecoratorConfig(['first gate'],
                                                   east_west = wall_37,
                                                   east_north = wall_13,
                                                   east_south = wall_35,
                                                   west_north = wall_17,
                                                   west_south = wall_57,
                                                   north_south = wall_15,
                                                   east_west_north = wall_137,
                                                   east_west_south = wall_357,
                                                   east_north_south = wall_135,
                                                   west_north_south = wall_157,
                                                   four_way = wall_1357,
                                                   wall = wall_constructed)

    wall_direction_builder = DirectionalWallDecorator(wall_direction_config)

    floor = surface_manager.add_icon('catacombs_floor', ':catacombs_floor.png', ' ')
    floor1 = surface_manager.add_icon('catacombs_floor_1', ':catacombs_floor_1.png', ' ')
    floor3 = surface_manager.add_icon('catacombs_floor_3', ':catacombs_floor_3.png', ' ')
    floor5 = surface_manager.add_icon('catacombs_floor_5', ':catacombs_floor_5.png', ' ')
    floor7 = surface_manager.add_icon('catacombs_floor_7', ':catacombs_floor_7.png', ' ')
    floor13 = surface_manager.add_icon('catacombs_floor_13', ':catacombs_floor_13.png', ' ')
    floor15 = surface_manager.add_icon('catacombs_floor_15', ':catacombs_floor_15.png', ' ')
    floor17 = surface_manager.add_icon('catacombs_floor_17', ':catacombs_floor_17.png', ' ')
    floor35 = surface_manager.add_icon('catacombs_floor_35', ':catacombs_floor_35.png', ' ')
    floor37 = surface_manager.add_icon('catacombs_floor_37', ':catacombs_floor_37.png', ' ')
    floor57 = surface_manager.add_icon('catacombs_floor_57', ':catacombs_floor_57.png', ' ')
    floor135 = surface_manager.add_icon('catacombs_floor_135', ':catacombs_floor_135.png', ' ')
    floor137 = surface_manager.add_icon('catacombs_floor_137', ':catacombs_floor_137.png', ' ')
    floor157 = surface_manager.add_icon('catacombs_floor_157', ':catacombs_floor_157.png', ' ')
    floor357 = surface_manager.add_icon('catacombs_floor_357', ':catacombs_floor_357.png', ' ')
    floor1357 = surface_manager.add_icon('catacombs_floor_1357', ':catacombs_floor_1357.png', ' ')

    floor_config = FloorBuilderDecoratorConfig([],
                                               single = floor,
                                               north = floor1,
                                               east = floor3,
                                               south = floor5,
                                               west = floor7,
                                               north_east = floor13,
                                               north_south = floor15,
                                               north_west = floor17,
                                               east_south = floor35,
                                               east_west = floor37,
                                               south_west = floor57,
                                               north_east_south = floor135,
                                               north_east_west = floor137,
                                               north_south_west = floor157,
                                               east_south_west = floor357,
                                               fourway = floor1357,
                                               floor = tile_floor)
    floor_builder = FloorBuilderDecorator(floor_config)

    torches_tile_f0 = surface_manager.add_icon('catacombs_torches_f0', ':wall_torches_f0.png', '¤')
    torches_tile_f1 = surface_manager.add_icon('catacombs_torches_f1', ':wall_torches_f1.png', '¤')
    torch_tile_f0 = surface_manager.add_icon('catacombs_torch_f0', ':wall_torch_f0.png', '¤')
    torch_tile_f1 = surface_manager.add_icon('catacombs_torch_f1', ':wall_torch_f1.png', '¤')

    torch_ornamenter_config = WallOrnamentDecoratorConfig(
                                                ['first gate'],
                                                wall_tile = wall_37,
                                                ornamentation = [(torch_tile_f0, torch_tile_f1),
                                                                 (torches_tile_f0, torches_tile_f1)],
                                                rng = rng,
                                                rate = 10)
    torch_ornamenter = WallOrnamentDecorator(torch_ornamenter_config)

    aggregate_decorator_config = AggregateDecoratorConfig(
                                                ['first gate'],
                                                [wallbuilder,
                                                 wall_direction_builder,
                                                 floor_builder,
                                                 torch_ornamenter])

    decorators = [AggregateDecorator(aggregate_decorator_config)]

    item_adder_config = ItemAdderConfiguration(['first gate'])

    item_adders = [ItemAdder(item_generator,
                            item_adder_config,
                            rng)]

    creatures_upper = CreatureAdderConfiguration(['first gate'])
    creatures_upper.add_creature(min_amount = 6,
                                 max_amount = 12,
                                 name = 'rat')
    creatures_upper.add_creature(min_amount = 0,
                                 max_amount = 2,
                                 name = 'fire beetle')
    creatures_upper.add_creature(min_amount = 0,
                                 max_amount = 1,
                                 name = 'skeleton warrior')

    creature_adders = [CreatureAdder(creature_generator,
                                    creatures_upper,
                                    rng)]

    portal_adder_configurations = [PortalAdderConfiguration(
                                        icons = (stairs_up,
                                                 stairs_down),
                                        level_type = 'first gate',
                                        location_type = 'room',
                                        chance = 100,
                                        new_level = 'lower catacombs',
                                        unique = True)]

    level_context = LevelContext(size = level_size,
                                 floor_type = floor_empty,
                                 wall_type = wall_natural,
                                 empty_floor = floor_empty,
                                 empty_wall = wall_empty,
                                 level_types = ['first gate'])

    config = (LevelConfiguration()
                    .with_rooms(room_generators)
                    .with_partitioners(level_partitioners)
                    .with_decorators(decorators)
                    .with_items(item_adders)
                    .with_creatures(creature_adders)
                    .with_portals(portal_adder_configurations)
                    .with_contexts([level_context])
                    .build())

    return config

def init_creatures(context):
    """
    Initialise creatures''

    :returns: list of configuration items
    :rtype: [CreatureConfiguration]
    """
    surface_manager = context.surface_manager
    config = []

    return config



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module for healing
"""
from pyherc.aspects import log_debug
from pyherc.data.effects.effect import Effect
from pyherc.events import HealAddedEvent, HealEndedEvent, HealTriggeredEvent


class Heal(Effect):
    """
    Class representing effects of healing
    """
    @log_debug
    def __init__(self, duration, frequency, tick, healing, target,
                 icon, title, description):
        """
        Default constructor
        """
        super().__init__(duration=duration,
                         frequency=frequency,
                         tick=tick,
                         icon=icon,
                         title=title,
                         description=description)

        self.healing = healing
        self.target = target
        self.effect_name = 'heal'

    @log_debug
    def do_trigger(self, dying_rules):
        """
        Triggers effects of the healing
        """
        self.target.hit_points = self.target.hit_points + self.healing

        if self.target.hit_points > self.target.max_hp:
            self.target.hit_points = self.target.max_hp

        self.target.raise_event(HealTriggeredEvent(target=self.target,
                                                   healing=self.healing))

    @log_debug
    def get_add_event(self):
        """
        Get event describing adding of this effect

        :returns: event describing adding of this effect
        :rtype: Event
        """
        return HealAddedEvent(target=self.target,
                              effect=self)

    @log_debug
    def get_removal_event(self):
        """
        Get event describing removal of this event

        :return: event describing removal of this event
        :rtype: Event
        """
        return HealEndedEvent(target=self.target,
                              effect=self)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
module for LevelGeneratorFactoryConfig
"""
from pyherc.generators.level.config import LevelGeneratorFactoryConfig


class LevelConfiguration():
    """
    DSL for configuring levels
    """

    def __init__(self):
        """
        Default constructor
        """
        super().__init__()
        self.room_generators = None
        self.level_partitioners = None
        self.decorators = None
        self.item_adders = None
        self.creature_adders = None
        self.portal_adder_configurations = None
        self.contexts = None

    def with_contexts(self, contexts):
        """
        Set level contexts which specify size and tiles to be used
        """
        self.contexts = contexts
        return self

    def with_rooms(self, rooms):
        """
        Add rooms to configuration

        :param rooms: rooms to use
        :type rooms: [RoomGenerator]
        """
        self.room_generators = rooms
        return self

    def with_partitioners(self, partitioners):
        """
        Add partitioners to configuration

        :param partitioners: partitioners to use
        :type partitioners: [LevelPartitioner]
        """
        self.level_partitioners = partitioners
        return self

    def with_decorators(self, decorators):
        """
        Add decorators to configuration

        :param decorators: decorators to use
        :type decorators: [LevelDecorator]
        """
        self.decorators = decorators
        return self

    def with_items(self, items):
        """
        Add items to configuration

        :param items: items to use
        :type items: [ItemAdder]
        """
        self.item_adders = items
        return self

    def with_creatures(self, creatures):
        """
        Add creatures to configuration

        :param creatures: creatures to use
        :type creatures: [CreatureAdder]
        """
        self.creature_adders = creatures
        return self

    def with_portals(self, portals):
        """
        Add portals to configuration

        :param portals: portals to use
        :type portals: [PortalAdder]
        """
        self.portal_adder_configurations = portals
        return self

    def build(self):
        """
        Build configuration

        :returns: configuration for level generator factory
        :rtype: LevelGeneratorFactoryConfig
        """
        return LevelGeneratorFactoryConfig(
            room_generators=self.room_generators,
            level_partitioners=self.level_partitioners,
            decorators=self.decorators,
            item_adders=self.item_adders,
            creature_adders=self.creature_adders,
            portal_adder_configurations=self.portal_adder_configurations,
            contexts=self.contexts)


class LevelContext():
    """
    Context for level generation
    """
    def __init__(self, size, floor_type, wall_type, empty_floor, empty_wall,
                 level_types):
        """
        Default constructor

        :param size: size of the level
        :type size: (int, int)
        :param floor_type: initial floor type to use
        :type floor_type: int
        :param wall_type: initial wall type to use
        :type wall_type: int
        :param empty_floor: floor tile to be considered empty
        :type empty_floor: int
        :param empty_wall: wall tile to be considered empty
        :type empty_wall: int
        """
        self.size = size
        self.floor_type = floor_type
        self.wall_type = wall_type
        self.empty_floor = empty_floor
        self.empty_wall = empty_wall
        self.level_types = level_types



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes for perception events
"""
from pyherc.events.event import Event


class NoticeEvent(Event):
    """
    Event that can be used to indicate character noticing something interesting

    .. versionadded:: 0.8
    """
    def __init__(self, character, target):
        """
        Default constructor
        """
        super().__init__(event_type='notice',
                         level=character.level,
                         location=character.location,
                         affected_tiles=[])

        self.character = character
        self.target = target

    def get_description(self, point_of_view):
        """
        Description of the event

        :param point_of_view: point of view for description
        :type point_of_view: Character
        :returns: description of the event
        :rtype: string
        """
        if point_of_view == self.character:
            description = 'You notice {0}'.format(self.target.name)
        elif point_of_view == self.target:
            description = '{0} notices you'.format(self.character.name)
        else:
            description = '{0} notices {1}'.format(self.character.name,
                                                   self.target.name)

        return description


class LoseFocusEvent(Event):
    """
    Event that can be used to indicate losing focus

    .. versionadded:: 0.8
    """
    def __init__(self, character):
        """
        Default constructor
        """
        super().__init__(event_type='lose focus',
                         level=character.level,
                         location=character.location,
                         affected_tiles=[])

        self.character = character

    def get_description(self, point_of_view):
        """
        Description of the event

        :param point_of_view: point of view for description
        :type point_of_view: Character
        :returns: description of the event
        :rtype: string
        """
        if point_of_view == self.character:
            description = 'You lose focus'
        else:
            description = '{0} loses focus'.format(self.character.name)

        return description



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Factory related classes are defined here
"""
import logging


class SubActionFactory():
    """
    Factory to handle concrete creation of actions
    """
    def __init__(self, effect_factory=None):
        """
        Constructor for this factory

        Args:
            effect_factory: Initialised EffectsFactory
        """
        super().__init__()
        self.action_type = 'default'
        self.logger = logging.getLogger('pyherc.rules.factory.SubActionFactory')  # noqa
        self.factories = []
        self.effect_factory = effect_factory

    def get_sub_factory(self, parameters):
        """
        Get sub factory to handle parameters

        Args:
            parameters: Parameters to use for searching the factory
        """
        subs = [x for x in self.factories if x.can_handle(parameters)]

        if len(subs) == 1:
            return subs[0]
        else:
            return None

    def can_handle(self, parameters):
        """
        Can this factory process these parameters

        Args:
            parameters: Parameters to check

        Returns:
            True if factory is capable of handling parameters
        """
        return self.action_type == parameters.action_type

    def get_action(self, parameters):
        """
        Create an action

        Args:
            parameters: Parameters used to control action creation
        """
        sub = self.get_sub_factory(parameters)
        return sub.get_action(parameters)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes for configuring level generation
"""
from pyherc.aspects import log_debug


class LevelGeneratorFactoryConfig():
    """
    Class to configure level generator
    """
    @log_debug
    def __init__(self, room_generators, level_partitioners,
                 decorators, item_adders, creature_adders,
                 portal_adder_configurations, contexts):
        """
        Default constructor
        """
        super().__init__()
        self.room_generators = room_generators
        self.level_partitioners = level_partitioners
        self.decorators = decorators
        self.item_adders = item_adders
        self.creature_adders = creature_adders
        self.portal_adder_configurations = portal_adder_configurations
        self.contexts = contexts



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes for generating corridors
"""

from pyherc.generators.level.partitioners.section import Connection


class CorridorGenerator():
    """
    Class for making simple corridors
    """
    def __init__(self, start_point, end_point, wall_tile, floor_tile):
        """
        Default constructor

        :param start_point: starting connection
        :type start_point: Connection
        :param end_point: ending connection
        :type end_point: Connection
        :param wall_tile: ID of wall tile to place
        :type wall_tile: integer
        :param floor_tile: ID of floor tile to place
        :type floor_tile: integer
        """
        self.start_point = start_point
        self.end_point = end_point
        self.wall_tile = wall_tile
        self.floor_tile = floor_tile

    def generate(self):
        """
        Carves corridor from start_point to end_point
        """
        if self.start_point.location[1] == self.end_point.location[1]:
            self.__carve_horizontal(self.start_point, self.end_point)
        elif self.start_point.location[0] == self.end_point.location[0]:
            self.__carve_vertical(self.start_point, self.end_point)
        elif self.start_point.direction in ("left", "right"):
            self.__carve_horizontal_bend(self.start_point, self.end_point)
        elif self.start_point.direction in ("up", "down"):
            self.__carve_vertical_bend(self.start_point, self.end_point)

    def __carve_horizontal(self, start_point, end_point):
        """
        Special case, carving is done in straigth horizontal line

        :param start_point: starting connection
        :type start_point: Connection
        :param end_point: ending connection
        :type end_point: Connection
        """
        if start_point.location[0] > end_point.location[0]:
            start_x = end_point.location[0]
            end_x = start_point.location[0]
        else:
            start_x = start_point.location[0]
            end_x = end_point.location[0]

        y_loc = start_point.location[1]
        section = start_point.section

        for x_loc in range(start_x, end_x):
            section.set_wall((x_loc, y_loc), self.wall_tile, 'corridor')
            section.set_floor((x_loc, y_loc), self.floor_tile, None)

        section.set_wall(end_point.location, self.wall_tile, 'corridor')
        section.set_wall(start_point.location, self.wall_tile, 'corridor')
        section.set_floor(end_point.location, self.floor_tile, None)
        section.set_floor(start_point.location, self.floor_tile, None)

    def __carve_horizontal_bend(self, start_point, end_point):
        """
        Carve corridor in three sections (horizontal, vertical, horizontal)

        :param start_point: starting connection
        :type start_point: Connection
        :param end_point: ending connection
        :type end_point: Connection
        """
        middle_x = abs(start_point.location[0] - end_point.location[0]) // 2

        if start_point.location[0] < end_point.location[0]:
            middle_x = middle_x + start_point.location[0]
        else:
            middle_x = middle_x + end_point.location[0]

        middle_start = Connection(None,
                                  (middle_x, start_point.location[1]),
                                  None,
                                  self.start_point.section)

        middle_end = Connection(None,
                                (middle_x, end_point.location[1]),
                                None,
                                self.start_point.section)

        self.__carve_horizontal(start_point, middle_start)
        self.__carve_vertical(middle_start, middle_end)
        self.__carve_horizontal(middle_end, end_point)

    def __carve_vertical(self, start_point, end_point):
        """
        Special case, carving is done in straigth vertical line

        :param start_point: starting connection
        :type start_point: Connection
        :param end_point: ending connection
        :type end_point: Connection
        """
        if start_point.location[1] > end_point.location[1]:
            start_y = end_point.location[1]
            end_y = start_point.location[1]
        else:
            start_y = start_point.location[1]
            end_y = end_point.location[1]

        x_loc = start_point.location[0]
        section = start_point.section

        for y_loc in range(start_y, end_y):
            section.set_wall((x_loc, y_loc), self.wall_tile, 'corridor')
            section.set_floor((x_loc, y_loc), self.floor_tile, None)

        section.set_wall(end_point.location, self.wall_tile, 'corridor')
        section.set_wall(start_point.location, self.wall_tile, 'corridor')
        section.set_floor(end_point.location, self.floor_tile, None)
        section.set_floor(start_point.location, self.floor_tile, None)

    def __carve_vertical_bend(self, start_point, end_point):
        """
        Carve corridor in three sections (vertical, horizontal,vertical)

        :param start_point: starting connection
        :type start_point: Connection
        :param end_point: ending connection
        :type end_point: Connection
        """
        middle_y = abs(start_point.location[1] - end_point.location[1]) // 2

        if start_point.location[1] < end_point.location[1]:
            middle_y = middle_y + start_point.location[1]
        else:
            middle_y = middle_y + end_point.location[1]

        middle_start = Connection(None,
                                  (start_point.location[0], middle_y),
                                  None,
                                  self.start_point.section)

        middle_end = Connection(None,
                                (end_point.location[0], middle_y),
                                None,
                                self.start_point.section)

        self.__carve_vertical(start_point, middle_start)
        self.__carve_horizontal(middle_start, middle_end)
        self.__carve_vertical(middle_end, end_point)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module defining classes related to damage
"""
from pyherc.aspects import log_debug


class Damage():
    """
    Damage done in attack
    """
    @log_debug
    def __init__(self, damage):
        """
        Default constructor
        """
        super().__init__()
        self.__damage = damage
        self.damage_inflicted = 0

    @log_debug
    def apply_damage(self, target):
        """
        Applies damage to target
        :param target: target to damage
        :type target: Character
        """
        for damage in self.__damage:
            damage_type = damage[1]

            matching_modifiers = [x for x in target.get_effects()
                                  if x.effect_name == 'damage modifier'
                                  and x.damage_type == damage_type]

            self.damage_inflicted = (self.damage_inflicted +
                                     damage[0] +
                                     sum(x.modifier for x
                                         in matching_modifiers))

            if self.damage_inflicted < 1:
                self.damage_inflicted = 0

        if target.inventory.armour is not None:
            armour = target.inventory.armour.armour_data.damage_reduction
        else:
            armour = 0

        if armour < self.damage_inflicted:
            self.damage_inflicted = self.damage_inflicted - armour
        else:
            if armour < self.damage_inflicted * 2:
                self.damage_inflicted = 1
            else:
                self.damage_inflicted = 0

        target.hit_points = target.hit_points - self.damage_inflicted

    @log_debug
    def __get_damage(self):
        """
        Total damage caused

        :returns: total damage caused
        :rtype: int
        """
        return sum(x[0] for x in self.__damage)

    @log_debug
    def __get_damage_types(self):
        """
        Types of damage caused

        :return: types of damage caused
        :rtype: [string]
        """
        return [x[1] for x
                in self.__damage]

    damage = property(__get_damage)
    damage_types = property(__get_damage_types)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Configuration for pyherc
"""
import random
from functools import partial

from pyherc.generators import (generate_creature, EffectsFactory,
                               ItemConfigurations, ItemGenerator,
                               SpellGenerator)
from pyherc.generators.level.config import LevelGeneratorFactoryConfig
from pyherc.generators.level.generator import LevelGeneratorFactory
from pyherc.generators.level.portals import PortalAdderFactory
from pyherc.rules import Dying, RulesEngine
from pyherc.rules.combat import RangedCombatFactory
from pyherc.rules.combat.factories import (AttackFactory, MeleeCombatFactory,
                                           UnarmedCombatFactory)
from pyherc.rules.consume.factories import DrinkFactory
from pyherc.rules.inventory.equip import EquipFactory
from pyherc.rules.inventory.factories import (DropFactory, InventoryFactory,
                                              PickUpFactory)
from pyherc.rules.inventory.unequip import UnEquipFactory
from pyherc.rules.magic import SpellCastingFactory
from pyherc.rules.moving.factories import MoveFactory, WalkFactory
from pyherc.rules.public import ActionFactory
from pyherc.rules.waiting import WaitFactory


class Configuration():
    """
    Configuration object for Herculeum
    """
    def __init__(self, model):
        """
        Default constructor

        Args:
            model: Model to register with factories
        """
        super().__init__()
        self.resolution = None
        self.full_screen = None
        self.caption = None
        self.action_factory = None
        self.item_generator = None
        self.creature_generator = None
        self.player_generator = None
        self.level_generator_factory = None
        self.level_size = None
        self.model = model
        self.rng = random.Random()
        self.rules_engine = None

    def initialise(self, context):
        """
        Initialises configuration
        """
        self.level_size = (79, 20)

        self.initialise_generators(context)
        self.initialise_level_generators(context)
        self.initialise_factories(context)

    def initialise_factories(self, context):
        """
        Initialises action factory, sub factories and various generators
        """
        dying_rules = Dying()

        walk_factory = WalkFactory(self.level_generator_factory,
                                   dying_rules)
        move_factory = MoveFactory(walk_factory)

        effect_factory = EffectsFactory()

        configurators = self.get_configurators(context.config_package,
                                               'init_effects')

        for configurator in configurators:
            effects = configurator(context)
            for effect in effects:
                effect_factory.add_effect(effect[0], effect[1])

        unarmed_combat_factory = UnarmedCombatFactory(effect_factory,
                                                      dying_rules)
        melee_combat_factory = MeleeCombatFactory(effect_factory,
                                                  dying_rules)
        ranged_combat_factory = RangedCombatFactory(effect_factory,
                                                    dying_rules)
        attack_factory = AttackFactory([unarmed_combat_factory,
                                        melee_combat_factory,
                                        ranged_combat_factory])

        drink_factory = DrinkFactory(effect_factory,
                                     dying_rules)

        inventory_factory = InventoryFactory([PickUpFactory(),
                                              DropFactory(),
                                              EquipFactory(),
                                              UnEquipFactory()])

        wait_factory = WaitFactory()

        spell_factory = SpellGenerator()

        spell_casting_factory = SpellCastingFactory(spell_factory,
                                                    effect_factory,
                                                    dying_rules)

        self.action_factory = ActionFactory(self.model,
                                            [move_factory,
                                             attack_factory,
                                             drink_factory,
                                             inventory_factory,
                                             wait_factory,
                                             spell_casting_factory])

        self.rules_engine = RulesEngine(self.action_factory,
                                        dying_rules)

    def get_creature_config(self, context):
        """
        Load creature configuration

        :param level_config: namespace of configurations
        :type level_config: Package
        :returns: configuration for creatures
        :rtype: CreatureConfigurations
        """
        config = {}

        configurators = self.get_configurators(context.config_package,
                                               'init_creatures')

        for configurator in configurators:
            creatures = configurator(context)
            for creature in creatures:
                config[creature['\ufdd0:name']] = creature

        return config

    def get_player_config(self, context):
        """
        Load player configuration

        :param level_config: namespace of configurations
        :type level_config: Package
        :returns: configuration for player characters
        :rtype: CreatureConfigurations

        .. versionadded:: 0.8
        """
        config = {}

        configurators = self.get_configurators(context.config_package,
                                               'init_players')

        for configurator in configurators:
            creatures = configurator(context)
            for creature in creatures:
                config[creature['\ufdd0:name']] = creature

        return config

    def get_item_config(self, context):
        """
        Load item configuration

        :param level_config: namespace of configurations
        :type level_config: Package
        :returns: configuration for items
        :rtype: ItemConfigurations
        """
        config = ItemConfigurations(self.rng)

        configurators = self.get_configurators(context.config_package,
                                               'init_items')

        for configurator in configurators:
            items = configurator(context)
            for item in items:
                config.add_item(item)

        return config

    def get_configurators(self, location, function_name):
        """
        Get functions to configure given sub-system
        """

        config_modules = map(lambda x: getattr(location, x),
                             filter(lambda x: x[0] != '_',
                                    dir(location)))
        configurators = map(lambda x: getattr(x, function_name),
                            filter(lambda y: hasattr(y, function_name),
                                   config_modules))

        return configurators

    def initialise_generators(self, context):
        """
        Initialise generators
        """
        self.item_generator = ItemGenerator(self.get_item_config(context))

        self.creature_generator = partial(generate_creature,
                                          self.get_creature_config(context),
                                          self.model,
                                          self.item_generator,
                                          self.rng)

        self.player_generator = partial(generate_creature,
                                        self.get_player_config(context),
                                        self.model,
                                        self.item_generator,
                                        self.rng)

        self.player_classes = self.get_player_config(context)

    def initialise_level_generators(self, context):
        """
        Initialise level generators

        :param level_config: module containing level configurations
        :type level_config: module
        """
        config = LevelGeneratorFactoryConfig([],
                                             [],
                                             [],
                                             [],
                                             [],
                                             [],
                                             [])

        configurators = self.get_configurators(context.config_package,
                                               'init_level')

        for configurator in configurators:
            self.extend_configuration(config,
                                      configurator(self.rng,
                                                   self.item_generator,
                                                   self.creature_generator,
                                                   self.level_size,
                                                   context))

        portal_adder_factory = PortalAdderFactory(
            config.portal_adder_configurations,
            self.rng)

        self.level_generator_factory = LevelGeneratorFactory(
            portal_adder_factory,
            config,
            self.rng)

    def extend_configuration(self, config, new_config):
        """
        Sums two configurations together

        Args:
            config: config to extend
            new_config: items to add to configuration
        """
        config.room_generators.extend(new_config.room_generators)
        config.level_partitioners.extend(new_config.level_partitioners)
        config.decorators.extend(new_config.decorators)
        config.item_adders.extend(new_config.item_adders)
        config.creature_adders.extend(new_config.creature_adders)
        config.portal_adder_configurations.extend(
            new_config.portal_adder_configurations)
        config.contexts.extend(new_config.contexts)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Public interface for wait actions
"""

from pyherc.aspects import log_debug, log_info
from pyherc.data import Duration
from pyherc.rules.public import ActionParameters


@log_info
def wait(character, action_factory):
    """
    Wait for a bit

    :param character: character doing waiting
    :type character: Character
    :param action_factory: factory to create actions
    :type action_factory: ActionFactory

    .. versionadded:: 0.10
    """
    action = action_factory.get_action(WaitParameters(character,
                                                      Duration.normal))
    if action.is_legal():
        action.execute()


class WaitParameters(ActionParameters):
    """
    Class for controlling waiting

    .. versionadded:: 0.10
    """
    @log_debug
    def __init__(self, character, time_to_wait):
        """
        Construct wait parameters

        :param character: character to wait
        :type character: Character
        :param time_to_wait: amount of ticks to wait
        :type time_to_wait: int
        """
        super().__init__()

        self.action_type = 'wait'
        self.character = character
        self.time_to_wait = time_to_wait



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module defining unarmed attack related objcts

Classes:
UnarmedToHit
UnarmedDamage
"""

from pyherc.aspects import log_debug
from pyherc.rules.combat.action import ToHit


class UnarmedToHit(ToHit):
    """
    Class to perform to hit calculations in unarmed combat
    """
    @log_debug
    def __init__(self, attacker, target, random_number_generator):
        """
        Default constructor

        Args:
            attacker: Character doing the attack
            target: Character being attacked
            random_number_generator: Random number generator
        """
        super().__init__(attacker, target, random_number_generator)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes for poison events
"""
from pyherc.events.event import Event


class DamageTriggeredEvent(Event):
    """
    Event that can be used to relay information about damage being triggered

    .. versionadded:: 0.8
    """
    def __init__(self, target, damage, damage_type):
        """
        Default constructor

        :param target: character who is suffering of damage
        :type target: Character
        :param damage: amount of damage suffered
        :type damage: int
        :param damage_type: type of damage
        :type damage_type: string
        """
        super().__init__(event_type='damage triggered',
                         level=target.level,
                         location=target.location,
                         affected_tiles=[])

        self.target = target
        self.damage = damage
        self.damage_type = damage_type

    def get_description(self, point_of_view):
        """
        Description of the event

        :param point_of_view: point of view for description
        :type point_of_view: Character
        :returns: description of the event
        :rtype: string
        """
        if point_of_view == self.target:
            description = 'You suffer from {0} damage ({1} points of damage)'.format(self.damage_type,  # noqa
                                                                                     self.damage)  # noqa
        else:
            description = '{0} suffers from {1} damage ({2} points of damage)'.format(self.target.name,  # noqa
                                                                               self.damage_type,  # noqa
                                                                               self.damage)  # noqa

        return description


class DamageAddedEvent(Event):
    """
    Event raised when character has been cast damage

    .. versionadded:: 0.8
    """
    def __init__(self, target, effect):
        """
        Default constructor

        :param target: target of the event
        :type target: Character
        :param effect: effect being added
        :type effect: Damage
        """
        super().__init__(event_type='damage started',
                         level=target.level,
                         location=target.location,
                         affected_tiles=[])

        self.target = target
        self.effect = effect

    def get_description(self, point_of_view):
        """
        Description of the event

        :param point_of_view: point of view for description
        :type point_of_view: Character
        :returns: description of the event
        :rtype: string
        """
        if point_of_view == self.target:
            description = 'You got hit by {0} damage'.format(self.effect.damage_type)  # noqa
        else:
            description = '{0} got hit by {1} damage'.format(self.target.name,
                                                             self.effect.damage_type)  # noqa

        return description


class DamageEndedEvent(Event):
    """
    Event to signal that damage effect is over
    """
    def __init__(self, target, effect):
        """
        Default constructor

        :param actor: character not suffering from damage anymore
        :type actor: Character
        :param effect: effect being removed
        :type effect: Damage
        """
        super().__init__(event_type='damage ended',
                         level=target.level,
                         location=target.location,
                         affected_tiles=[])

        self.target = target
        self.effect = effect

    def get_description(self, point_of_view):
        """
        Description of the event

        :param point_of_view: point of view for description
        :type point_of_view: Character
        :returns: description of the event
        :rtype: string
        """
        if point_of_view == self.target:
            description = 'You are no longer suffering of {0} damage'.format(self.effect.damage_type)  # noqa
        else:
            description = '{0} is no longer suffering {1} damage'.format(self.target.name,  # noqa
                                                                         self.effect.damage_type)  # noqa

        return description



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

# flake8: noqa

"""
Package for rules of the game
"""
from pyherc.rules.moving.interface import move, is_move_legal
from pyherc.rules.magic.interface import cast, gain_domain
from pyherc.rules.inventory.interface import pick_up, drop_item
from pyherc.rules.inventory.interface import equip, unequip
from pyherc.rules.combat.interface import attack
from pyherc.rules.consume.interface import drink
from pyherc.rules.waiting.interface import wait

from .public import ActionFactory
from .public import ActionParameters
from .ending import Dying
from .engine import RulesEngine



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
module for configuring lair of crimson jaw
"""
from herculeum.ai import SkeletonWarriorAI
from pyherc.config.dsl import LevelConfiguration, LevelContext
from pyherc.data.effects import EffectHandle
from pyherc.generators import (creature_config, inventory_config,
                               ItemConfiguration, WeaponConfiguration)
from pyherc.generators.level.creatures import (CreatureAdder,
                                               CreatureAdderConfiguration)
from pyherc.generators.level.decorator import (AggregateDecorator,
                                               AggregateDecoratorConfig,
                                               DirectionalWallDecorator,
                                               DirectionalWallDecoratorConfig,
                                               FloorBuilderDecorator,
                                               FloorBuilderDecoratorConfig,
                                               ReplacingDecorator,
                                               ReplacingDecoratorConfig,
                                               WallBuilderDecorator,
                                               WallBuilderDecoratorConfig)
from pyherc.generators.level.items import ItemAdder, ItemAdderConfiguration
from pyherc.generators.level.partitioners import GridPartitioner
from pyherc.generators.level.portals import PortalAdderConfiguration
from pyherc.generators.level.room import CrimsonLairGenerator
from pyherc.rules.constants import SLASHING_DAMAGE


def init_level(rng, item_generator, creature_generator, level_size, context):
    """
    Initialise lair of crimson jaw

    :returns: level configuration
    :rtype: LevelConfiguration
    """
    surface_manager = context.surface_manager

    floor_natural = 'natural floor'
    floor_rock = surface_manager.add_icon('floor_rock', ':rock_floor.png', '.')

    wall_empty = surface_manager.add_icon('empty wall', ':transparent.png', None)
    wall_natural = 'natural wall'
    wall_constructed = 'constructed wall'
    wall_ground = surface_manager.add_icon('wall_ground', ':ground_wall.png', ' ')
    wall_rock = surface_manager.add_icon('wall_rock', ':rock_wall.png', '#')
    stairs_down = surface_manager.add_icon('stairs_down', ':stairs_down.png', '>')
    stairs_up = surface_manager.add_icon('stairs_up', ':stairs_up.png', '<')

    room_generators = [CrimsonLairGenerator(floor_natural,
                                           wall_empty,
                                           ['crimson lair'],
                                           rng)]
    level_partitioners = [GridPartitioner(['crimson lair'],
                                           1,
                                           1,
                                           rng)]

    replacer_config = ReplacingDecoratorConfig(['crimson lair'],
                                    {floor_natural: floor_rock},
                                    {wall_natural: wall_ground,
                                     wall_constructed: wall_rock})
    replacer = ReplacingDecorator(replacer_config)

    wallbuilder_config = WallBuilderDecoratorConfig(['crimson lair'],
                                        {wall_natural: wall_constructed},
                                         wall_empty)
    wallbuilder = WallBuilderDecorator(wallbuilder_config)

    wall = surface_manager.add_icon('abyss_wall', ':abyss_wall.png', '#')
    wall_15 = surface_manager.add_icon('abyss_wall_15', ':abyss_wall_15.png', '#')
    wall_57 = surface_manager.add_icon('abyss_wall_57', ':abyss_wall_57.png', '#')
    wall_35 = surface_manager.add_icon('abyss_wall_35', ':abyss_wall_35.png', '#')
    wall_37 = surface_manager.add_icon('abyss_wall_37', ':abyss_wall_37.png', '#')
    wall_17 = surface_manager.add_icon('abyss_wall_17', ':abyss_wall_17.png', '#')
    wall_13 = surface_manager.add_icon('abyss_wall_13', ':abyss_wall_13.png', '#')
    wall_135 = surface_manager.add_icon('abyss_wall_135', ':abyss_wall_135.png', '#')
    wall_357 = surface_manager.add_icon('abyss_wall_357', ':abyss_wall_357.png', '#')
    wall_1357 = surface_manager.add_icon('abyss_wall_1357', ':abyss_wall_1357.png', '#')
    wall_137 = surface_manager.add_icon('abyss_wall_137', ':abyss_wall_137.png', '#')
    wall_157 = surface_manager.add_icon('abyss_wall_157', ':abyss_wall_157.png', '#')

    floor = surface_manager.add_icon('abyss_floor', ':abyss_floor.png', ' ')
    floor1 = surface_manager.add_icon('abyss_floor_1', ':abyss_floor_1.png', ' ')
    floor3 = surface_manager.add_icon('abyss_floor_3', ':abyss_floor_3.png', ' ')
    floor5 = surface_manager.add_icon('abyss_floor_5', ':abyss_floor_5.png', ' ')
    floor7 = surface_manager.add_icon('abyss_floor_7', ':abyss_floor_7.png', ' ')
    floor13 = surface_manager.add_icon('abyss_floor_13', ':abyss_floor_13.png', ' ')
    floor15 = surface_manager.add_icon('abyss_floor_15', ':abyss_floor_15.png', ' ')
    floor17 = surface_manager.add_icon('abyss_floor_17', ':abyss_floor_17.png', ' ')
    floor35 = surface_manager.add_icon('abyss_floor_35', ':abyss_floor_35.png', ' ')
    floor37 = surface_manager.add_icon('abyss_floor_37', ':abyss_floor_37.png', ' ')
    floor57 = surface_manager.add_icon('abyss_floor_57', ':abyss_floor_57.png', ' ')
    floor135 = surface_manager.add_icon('abyss_floor_135', ':abyss_floor_135.png', ' ')
    floor137 = surface_manager.add_icon('abyss_floor_137', ':abyss_floor_137.png', ' ')
    floor157 = surface_manager.add_icon('abyss_floor_157', ':abyss_floor_157.png', ' ')
    floor357 = surface_manager.add_icon('abyss_floor_357', ':abyss_floor_357.png', ' ')
    floor1357 = surface_manager.add_icon('abyss_floor_1357', ':abyss_floor_1357.png', ' ')

    floor_config = FloorBuilderDecoratorConfig([],
                                               single = floor,
                                               north = floor1,
                                               east = floor3,
                                               south = floor5,
                                               west = floor7,
                                               north_east = floor13,
                                               north_south = floor15,
                                               north_west = floor17,
                                               east_south = floor35,
                                               east_west = floor37,
                                               south_west = floor57,
                                               north_east_south = floor135,
                                               north_east_west = floor137,
                                               north_south_west = floor157,
                                               east_south_west = floor357,
                                               fourway = floor1357,
                                               floor = floor_natural)
    floor_builder = FloorBuilderDecorator(floor_config)

    wall_direction_config = DirectionalWallDecoratorConfig(['upper crypt'],
                                                   east_west = wall_37,
                                                   east_north = wall_13,
                                                   east_south = wall_35,
                                                   west_north = wall_17,
                                                   west_south = wall_57,
                                                   north_south = wall_15,
                                                   east_west_north = wall_137,
                                                   east_west_south = wall_357,
                                                   east_north_south = wall_135,
                                                   west_north_south = wall_157,
                                                   four_way = wall_1357,
                                                   wall = wall_constructed)

    wall_direction_builder = DirectionalWallDecorator(wall_direction_config)

    aggregate_decorator_config = AggregateDecoratorConfig(
                                                ['crimson lair'],
                                                [wallbuilder,
                                                wall_direction_builder,
                                                floor_builder,
                                                replacer])

    decorators = [AggregateDecorator(aggregate_decorator_config)]

    item_adder_config = ItemAdderConfiguration(['crimson lair'])
    item_adder_config.add_item(min_amount = 2,
                               max_amount = 4,
                               type = 'weapon',
                               location = 'room')
    item_adder_config.add_item(min_amount = 0,
                               max_amount = 2,
                               type = 'potion',
                               location = 'room')
    item_adder_config.add_item(min_amount = 1,
                               max_amount = 3,
                               type = 'food',
                               location = 'room')
    item_adders = [ItemAdder(item_generator,
                            item_adder_config,
                            rng)]

    creatures_upper = CreatureAdderConfiguration(['crimson lair'])
    creatures_upper.add_creature(min_amount = 1,
                                 max_amount = 1,
                                 name = 'crimson jaw')

    creature_adders = [CreatureAdder(creature_generator,
                                    creatures_upper,
                                    rng)]

    portal_adder_configurations = [PortalAdderConfiguration(
                                        icons = (stairs_down,
                                                 stairs_up),
                                        level_type = 'lower catacombs',
                                        location_type = 'room',
                                        chance = 10,
                                        new_level = 'crimson lair',
                                        unique = True)
                                        ]

    level_context = LevelContext(size = level_size,
                                 floor_type = 0,
                                 wall_type = wall_natural,
                                 empty_floor = 0,
                                 empty_wall = wall_empty,
                                 level_types = ['crimson lair'])

    config = (LevelConfiguration()
                    .with_rooms(room_generators)
                    .with_partitioners(level_partitioners)
                    .with_decorators(decorators)
                    .with_items(item_adders)
                    .with_creatures(creature_adders)
                    .with_portals(portal_adder_configurations)
                    .with_contexts([level_context])
                    .build())

    return config

def init_creatures(context):
    """
    Initialise creatures

    :returns: list of creature configurations
    :rtype: [CreatureConfiguration]
    """
    config = []
    surface_manager = context.surface_manager


    crimson_jaw_f0 = surface_manager.add_icon('crimson jaw_f0', ':crimson-jaw_f0.png', '&', ['bold', 'red'])
    crimson_jaw_f1 = surface_manager.add_icon('crimson jaw_f1', ':crimson-jaw_f1.png', '&', ['bold', 'red'])
    config.append(creature_config(name = 'crimson jaw',
                                  body = 12,
                                  finesse = 9,
                                  mind = 12,
                                  hp = 30,
                                  speed = 3,
                                  icons = (crimson_jaw_f0, crimson_jaw_f1),
                                  attack = 8,
                                  inventory = [inventory_config(
                                      item_name = 'whip of ashmque',
                                      min_amount = 0,
                                      max_amount = 1,
                                      probability = 100)],
                                  ai = SkeletonWarriorAI))

    return config

def init_items(context):
    """
    Initialise special items

    :returns: item configurations
    :rtype: [ItemConfiguration]
    """
    surface_manager = context.surface_manager
    config = []

    config.append(
                  ItemConfiguration(name = 'whip of ashmque',
                                    cost = 0,
                                    weight = 12,
                                    icons = [surface_manager.add_icon('whip of ashmque', ':ashmque.png', ')', ['bold', 'red'])],
                                    types = ['weapon',
                                             'one-handed',
                                             'melee',
                                             'exotic weapon'],
                                    rarity = 'artifact',
                                    effect_handles = [EffectHandle(
                                            trigger = 'on attack hit',
                                            effect = 'major fire damage',
                                            parameters = None,
                                            charges = 999999)],
                                    weapon_configration = WeaponConfiguration(
                                            damage = [(12, SLASHING_DAMAGE)],
                                            critical_range = 10,
                                            critical_damage = 12,
                                            weapon_class = 'exotic')))

    return config



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes for inventory events
"""
from pyherc.events.event import Event


class PickUpEvent(Event):
    """
    Event that can be used to relay information about item being picked up

    .. versionadded:: 0.5
    """
    def __init__(self, character, item):
        """
        Default constructor

        :param character: character picking up the item
        :type character: Character
        :param item: item being picked up
        :type item: Item
        """
        super().__init__(event_type='pick up',
                         level=character.level,
                         location=character.location,
                         affected_tiles=[])

        self.character = character
        self.item = item

    def get_description(self, point_of_view):
        """
        Description of the event

        :param point_of_view: point of view for description
        :type point_of_view: Character
        :returns: description of the event
        :rtype: string
        """
        if point_of_view == self.character:
            description = 'You pick up {0}'.format(self.item.name)
        else:
            description = '{0} picks up {1}'.format(self.character.name,
                                                    self.item.name)

        return description


class DropEvent(Event):
    """
    Event that can be used to relay information about item being dropped

    .. versionadded:: 0.5
    """
    def __init__(self, character, item):
        """
        Default constructor

        :param character: character dropping the item
        :type character: Character
        :param item: item being dropped
        :type item: Item
        """
        super().__init__(event_type='drop',
                         level=character.level,
                         location=character.location,
                         affected_tiles=[])

        self.character = character
        self.item = item

    def get_description(self, point_of_view):
        """
        Description of the event

        :param point_of_view: point of view for description
        :type point_of_view: Character
        :returns: description of the event
        :rtype: string
        """
        if point_of_view == self.character:
            description = 'You drop {0}'.format(self.item.name)
        else:
            description = '{0} drops {1}'.format(self.character.name,
                                                 self.item.name)

        return description


class EquipEvent(Event):
    """
    Event that can be used to relay information about item being equiped

    .. versionadded:: 0.8
    """
    def __init__(self, character, item):
        """
        Default constructor

        :param character: character equiping the item
        :type character: Character
        :param item: item being equipped
        :type item: Item
        """
        super().__init__(event_type='equip',
                         level=character.level,
                         location=character.location,
                         affected_tiles=[])

        self.character = character
        self.item = item

    def get_description(self, point_of_view):
        """
        Description of the event

        :param point_of_view: point of view for description
        :type point_of_view: Character
        :returns: description of the event
        :rtype: string
        """
        if point_of_view == self.character:
            description = 'You equip {0}'.format(self.item.name)
        else:
            description = '{0} equips {1}'.format(self.character.name,
                                                  self.item.name)

        return description


class UnEquipEvent(Event):
    """
    Event that can be used to relay information about item being unequiped

    .. versionadded:: 0.8
    """
    def __init__(self, character, item):
        """
        Default constructor

        :param character: character unequiping the item
        :type character: Character
        :param item: item being unequipped
        :type item: Item
        """
        super().__init__(event_type='unequip',
                         level=character.level,
                         location=character.location,
                         affected_tiles=[])

        self.character = character
        self.item = item

    def get_description(self, point_of_view):
        """
        Description of the event

        :param point_of_view: point of view for description
        :type point_of_view: Character
        :returns: description of the event
        :rtype: string
        """
        if point_of_view == self.character:
            description = 'You unequip {0}'.format(self.item.name)
        else:
            description = '{0} unequips {1}'.format(self.character.name,
                                                    self.item.name)

        return description



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module for checking end conditions
"""

from pyherc.aspects import log_debug, log_info
from pyherc.data.model import DIED_IN_DUNGEON, ESCAPED_DUNGEON
from pyherc.events import DeathEvent, DropEvent


class Dying():
    """
    Rules for actions happening when dying

    .. versionadded:: 0.6
    """
    @log_debug
    def __init__(self):
        """
        Default constructor
        """
        super().__init__()

    @log_debug
    def check_dying(self, character):
        """
        Check if character dies and process it
        """
        if not character.level:
            return

        if character.hit_points <= 0:
            for item in character.inventory:
                character.inventory.remove(item)
                character.level.add_item(item, character.location)
                character.raise_event(DropEvent(character,
                                                item))

            if character.inventory.weapon is not None:
                character.level.add_item(character.inventory.weapon,
                                         character.location)
                character.raise_event(DropEvent(character,
                                                item))
                character.inventory.weapon = None

            if character == character.model.player:
                character.model.end_condition = DIED_IN_DUNGEON

            character.raise_event(
                DeathEvent(deceased=character,
                           affected_tiles=character.location))
            character.level.remove_creature(character)

    @log_info
    def calculate_score(self, character):
        """
        Calculate score for character

        .. versionadded:: 0.8
        """
        end_condition = character.model.end_condition

        score = sum((item.cost for item in character.inventory))

        if end_condition == DIED_IN_DUNGEON:
            score = int(score * 0.75)
        elif end_condition == ESCAPED_DUNGEON:
            score = int(score * 1.25)

        return score



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module for moving
"""
from pyherc.rules import attack, is_move_legal, move


class MoveController():
    """
    Controller for moving around

    .. versionadded:: 0.9
    """
    def __init__(self, action_factory, rng):
        """
        Default constructor
        """
        super().__init__()

        self.action_factory = action_factory
        self.rng = rng

    def move_or_attack(self, character, direction, movement_mode):
        """
        Move or attack
        """
        level = character.level

        if is_move_legal(character,
                         direction,
                         movement_mode,
                         self.action_factory):
            move(character, direction, self.action_factory)
        elif direction != 9:
            loc = character.get_location_at_direction(direction)
            if level.get_creature_at(loc) is not None:
                attack(character,
                       direction,
                       self.action_factory,
                       self.rng)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Wait related factories
"""
from pyherc.aspects import log_debug, log_info
from pyherc.rules.factory import SubActionFactory
from pyherc.rules.waiting.action import WaitAction


class WaitFactory(SubActionFactory):
    """
    Factory for creating wait actions

    .. versionadded:: 0.10
    """
    @log_debug
    def __init__(self):
        """
        Constructor for this factory
        """
        super().__init__()
        self.action_type = 'wait'

    @log_info
    def get_action(self, parameters):
        """
        Create a wait action

        :param parameters: parameters used to control creation
        :type parameters: WaitParameters
        """
        return WaitAction(character=parameters.character,
                          time_to_wait=parameters.time_to_wait)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes for move events
"""
from pyherc.events.event import Event


class MoveEvent(Event):
    """
    Event that can be used to relay information about moving

    .. versionadded:: 0.4
    """
    def __init__(self, mover, old_location, direction, affected_tiles):
        """
        Default constructor

        :param mover: character moving around
        :type mover: Character
        :param affected_tiles: tiles affected by event
        :type affected_tiles: [(int, int)]
        """
        super().__init__(event_type='move',
                         level=mover.level,
                         location=mover.location,
                         affected_tiles=affected_tiles)

        self.mover = mover
        self.old_location = old_location
        self.direction = direction

    def get_description(self, point_of_view):
        """
        Description of the event

        :param point_of_view: point of view for description
        :type point_of_view: Character
        :returns: description of the event
        :rtype: string
        """
        if point_of_view == self.mover:
            description = 'You move'
        else:
            description = '{0} moves'.format(self.mover.name)

        return description



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module for spell book related classes
"""

from pyherc.aspects import log_debug, log_info


class SpellBook():
    """
    List of known spells

    .. versionadded:: 0.10
    """

    @log_info
    def __init__(self):
        """
        Default constructor
        """
        super().__init__()

        self.domains = {}
        self.spells = []

    @log_debug
    def add_domain_level(self, domain, level=None):
        """
        Add level to a spell domain

        :param domain: name of domain to learn
        :type domain: string
        :param level: amount to increment the level
        :type level: int

        .. note:: if level is None, current level is incremented by one
        """
        if not level:
            level = 1

        if domain in self.domains:
            self.domains[domain] = self.domains[domain] + level
        else:
            self.domains[domain] = level

    @log_debug
    def get_domain_level(self, domain):
        """
        Get current spell level of a given domain

        :param domain: name of the domain
        :type domain: String
        :returns: current level, 0 if domain is unknown
        :rtype: int

        .. versionadded:: 0.10
        """
        if domain in self.domains:
            return self.domains[domain]
        else:
            return 0

    @log_debug
    def add_spell_entry(self, entry):
        """
        Add spell entry into spellbook

        :param entry: entry to add
        :type entry: SpellEntry

        .. versionadded:: 0.10
        """
        self.spells.append(entry)

    @log_debug
    def get_known_spells(self):
        """
        Get a list of known spells

        :returns: list of known spells
        :rtype: [SpellEntry]
        """
        known_spells = []

        for domain, level in self.domains.items():
            known_spells.extend([spell for spell in self.spells
                                 if spell.level <= level
                                 and spell.domain == domain])

        return known_spells


class SpellEntry():
    """
    Entry in a spell book

    .. versionadded:: 0.10
    """
    def __init__(self):
        """
        Default constructor
        """
        super().__init__()

        self.domain = None
        self.level = None
        self.spell_name = None

    def __str__(self):
        """
        String representation of this spell entry
        """
        return '{0} ({1}:{2})'.format(self.spell_name,
                                      self.domain,
                                      self.level)

    def __repr__(self):
        """
        String representation of this spell entry
        """
        return self.__str__()



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Attack related factories are defined here
"""
from pyherc.aspects import log_debug, log_info
from pyherc.data.damage import Damage
from pyherc.data.geometry import get_adjacent_target_in_direction
from pyherc.rules.combat.action import AdditionalRules, AttackAction
from pyherc.rules.combat.melee import MeleeToHit
from pyherc.rules.combat.unarmed import UnarmedToHit
from pyherc.rules.factory import SubActionFactory


class AttackFactory(SubActionFactory):
    """
    Factory for constructing attack actions
    """

    @log_debug
    def __init__(self, factories):
        """
        Constructor for this factory
        """
        super().__init__()
        self.action_type = 'attack'

        if hasattr(factories, '__iter__'):
            self.factories = factories
        else:
            self.factories = []
            self.factories.append(factories)


class UnarmedCombatFactory():
    """
    Factory for producing unarmed combat actions
    """
    @log_debug
    def __init__(self, effect_factory, dying_rules):
        """
        Constructor for this factory
        """
        super().__init__()
        self.attack_type = 'unarmed'
        self.effect_factory = effect_factory
        self.dying_rules = dying_rules

    def __str__(self):
        return 'unarmed combat factory'

    @log_debug
    def can_handle(self, parameters):
        """
        Can this factory process these parameters

        :param parameters: parameters to check
        :type parameters: UnarmedAttackParameters
        :returns: true if factory is capable of handling parameters
        :rtype: boolean
        """
        return self.attack_type == parameters.attack_type

    @log_info
    def get_action(self, parameters):
        """
        Create a attack action

        :param parameters: parameters used to control attack creation
        :type parameters: UnarmedAttackParameters
        :returns: action that can be executed
        :rtype: AttackAction
        """
        attacker = parameters.attacker
        target = self.get_target(parameters)
        damage = [(attacker.get_attack(), 'crushing')]

        attack = AttackAction(
            attack_type='unarmed',
            to_hit=UnarmedToHit(attacker,
                                target,
                                parameters.random_number_generator),
            damage=Damage(damage=damage),
            attacker=attacker,
            target=target,
            effect_factory=self.effect_factory,
            dying_rules=self.dying_rules,
            additional_rules=AdditionalRules(attacker))

        return attack

    @log_debug
    def get_target(self, parameters):
        """
        Get target of the attack

        :param parameters: attack parameters
        :type parameters: UnarmedAttackParameters
        :returns: target character if found, otherwise None
        :rtype: AttackData
        """
        attacker = parameters.attacker
        direction = parameters.direction
        level = attacker.level

        return get_adjacent_target_in_direction(level,
                                                attacker.location,
                                                direction)


class MeleeCombatFactory():
    """
    Factory for producing melee combat actions
    """
    @log_debug
    def __init__(self, effect_factory, dying_rules):
        """
        Constructor for this factory
        """
        super().__init__()
        self.attack_type = 'melee'
        self.effect_factory = effect_factory
        self.dying_rules = dying_rules

    def __str__(self):
        return 'melee combat factory'

    @log_debug
    def can_handle(self, parameters):
        """
        Can this factory process these parameters

        :param parameters: parameters to check
        :type parameters: MeleeAttackParameters
        :returns: true if factory is capable of handling parameters
        :rtype: boolean
        """
        return self.attack_type == parameters.attack_type

    @log_info
    def get_action(self, parameters):
        """
        Create a attack action

        :param parameters: parameters used to control attack creation
        :type parameters: MeleeAttackParameters
        :returns: action that can be executed
        :rtype: AttackAction
        """
        attacker = parameters.attacker
        target = self.get_target(parameters)
        weapon = attacker.inventory.weapon
        weapon_data = weapon.weapon_data

        attack = AttackAction(
            attack_type='melee',
            to_hit=MeleeToHit(attacker,
                              target,
                              parameters.random_number_generator),
            damage=Damage(damage=weapon_data.damage),
            attacker=attacker,
            target=target,
            effect_factory=self.effect_factory,
            dying_rules=self.dying_rules,
            additional_rules=AdditionalRules(attacker))

        return attack

    @log_debug
    def get_target(self, parameters):
        """
        Get target of the attack

        :param parameters: parameters to control attack
        :type parameters: MeleeAttackParameters
        :returns: target character if found, otherwise None
        :rtype: AttackData
        """
        attacker = parameters.attacker
        direction = parameters.direction
        level = attacker.level

        return get_adjacent_target_in_direction(level,
                                                attacker.location,
                                                direction)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module for end screen
"""
from datetime import date


class EndScreen():
    """
    Screen for displaying end of the game results

    .. versionadded:: 0.9
    """
    def __init__(self, model, dying_rules, screen,
                 controller):
        """
        Default constructor
        """
        super(EndScreen, self).__init__()

        self.model = model
        self.dying_rules = dying_rules
        self.screen = screen.derwin(18, 40, 5, 20)
        self.controller = controller

    def show(self):
        """
        Display the screen
        """
        self.screen.clear()
        self.screen.border()

        self.screen.addstr(2, 2, 'Date: {0}'.format(date.today()))
        self.screen.addstr(3, 2, 'Score: {0}'.format(
                                        self.dying_rules.calculate_score(
                                                        self.model.player)))
        self.screen.addstr(5, 2, self.controller.get_end_description(
                                                    self.model.end_condition))

        self.screen.refresh()
        self.screen.getch()



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes for generating catacombs
"""
from pyherc.aspects import log_debug


class CrimsonLairGenerator():
    """
    Class for generating a catacomblike rooms
    """
    @log_debug
    def __init__(self, floor_tile, empty_tile, level_types, rng):
        """
        Default constructor

        :param floor_tile: id of the tile to use for floors
        :type floor_tile: integer
        :param empty_tile: id of the empty wall tile
        :type empty_tile: integer
        :param level_types: types of level this generator can be used
        :type level_types: [string]
        :param rng: random number generator
        :type rng: Random
        """
        super().__init__()
        self.floor_tile = floor_tile
        self.empty_tile = empty_tile
        self.room_width = None
        self.room_height = None
        self.level_types = level_types
        self.rng = rng

    @log_debug
    def generate_room(self, section):
        """
        Generate room

        :param section: section for generator to draw to
        :type section: Section
        """
        level_size = (section.width, section.height)
        for x_loc in range(1, level_size[0]):
            for y_loc in range(1, level_size[1]):
                section.set_floor((x_loc, y_loc),
                                  self.floor_tile,
                                  'room')
                section.set_wall((x_loc, y_loc),
                                 self.empty_tile,
                                 None)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module for curses user interface
"""
import curses

from herculeum.ui.controllers import StartGameController
from herculeum.ui.text.main_window import MainWindow


class CursesUserInterface():
    """
    Class for curses user interface

    .. versionadded:: 0.9
    """
    def __init__(self, application):
        """
        Default constructor
        """
        super().__init__()

        self.application = application
        self.splash_screen = None

        self.curses = curses.initscr()

        if curses.has_colors():
            curses.start_color()

        self.screen = self.curses.subwin(24, 80, 0, 0)

        self.screen.refresh()
        curses.noecho()
        curses.cbreak()
        curses.curs_set(0)

    def show_splash_screen(self):
        """
        Show splash screen
        """
        pass

    def show_main_window(self):
        """
        Show main window
        """
        main_window = MainWindow(self.application,
                                 self.application.surface_manager,
                                 self.screen,
                                 StartGameController(self.application.level_generator_factory,
                                                     self.application.creature_generator,
                                                     self.application.item_generator,
                                                     self.application.config.start_level))

        main_window.show_new_game()
        main_window.show_map_window()

        curses.echo()
        curses.nocbreak()
        curses.curs_set(1)
        curses.endwin()



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Port to issue actions
"""
from pyherc.rules import is_move_legal, move


class ActionsPort():
    """
    Port to issue actions with

    .. versionadded:: 0.10
    """

    def __init__(self, action_factory):
        """
        Default constructor

        :param action_factory: action factory to use
        :type action_factory: ActionFactory
        """
        super().__init__()

        self.action_factory = action_factory

    def move_character(self, character, direction):
        """
        Move a character to given direction

        :param character: character to move
        :type character: Character
        :param direction: direction to move to
        :type direction: int (1-8)
        """
        move(character=character,
             direction=direction,
             action_factory=self.action_factory)

    def is_move_legal(self, character, direction):
        """
        Check if character is allowed to move to given direction

        :param character: character to move
        :type character: Character
        :param direction: direction to move to
        :type direction: int (1-8)
        """
        return is_move_legal(character=character,
                             direction=direction,
                             movement_mode='walk',
                             action_factory=self.action_factory)



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module for damage modifiers
"""
from pyherc.aspects import log_debug
from pyherc.data.effects.effect import Effect


class DamageModifier(Effect):
    """
    Class representing damage modifier

    .. versionadded:: 0.7
    """
    @log_debug
    def __init__(self, modifier, damage_type, duration, frequency, tick,
                 icon, title, description):
        """
        Default constructor
        """
        super().__init__(duration=duration,
                         frequency=frequency,
                         tick=tick,
                         icon=icon,
                         title=title,
                         description=description)

        self.modifier = modifier
        self.damage_type = damage_type
        self.effect_name = 'damage modifier'
        self.multiple_allowed = True

    @log_debug
    def clone(self):
        """
        Temporary hack for tech day
        """
        return DamageModifier(self.modifier, self.damage_type,
                              self.duration, self.frequency,
                              self.tick, self.icon, self.title,
                              self.description)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module for character screen
"""
class CharacterScreen():
    """
    Character screen

    .. versionadded:: 0.9
    """
    def __init__(self, character, screen):
        """
        Default constructor
        """
        super(CharacterScreen, self).__init__()

        self.screen = screen.subwin(20, 75, 2, 2)
        self.character = character

    def show(self):
        """
        Show the dialog
        """
        self._draw_screen()
        chr(self.screen.getch())

    def _draw_screen(self):
        """
        Update the screen
        """
        self.screen.clear()
        self.screen.border()

        player = self.character

        self.screen.addstr(1, 1, 'hp: {0}/{1}'.format(player.hit_points,
                                                     player.max_hp))
        self.screen.addstr(2, 1, 'mana: 0/0')
        self.screen.addstr(1, 37, 'body: {0}'.format(player.body))
        self.screen.addstr(2, 37, 'mind: {0}'.format(player.mind))
        self.screen.addstr(3, 37, 'finesse: {0}'.format(player.finesse))

        self.screen.addstr(5, 1, 'Effects:')
        self.screen.addstr(5, 37, 'Skills:')
        for index, effect in enumerate(player.get_effects()):
            self.screen.addstr(6 + index, 1, effect.title)

        self.screen.refresh()



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes for death animations
"""
from .animation import Animation


class DeathAnimation(Animation):
    """
    Generic dying animation

    .. versionadded:: 0.12
    """
    def __init__(self, event):
        """
        Default constructor
        """
        super().__init__(event)

        self.deceased = event.deceased

    def trigger(self, ui):
        """
        Trigger this animation
        """
        glyphs = [x for x in ui.view.items()
                  if (hasattr(x, 'entity'))
                  and (x.entity == self.deceased)]

        for glyph in glyphs:
            ui.view.scene().removeItem(glyph)

        if self.deceased == ui.model.player:
            ui.EndScreenRequested.emit()



"""{-
Copyright (c) 2009 Pauli Henrikki Rikula

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
-}"""

class NotImplemented(Exception):
    pass



class HeapSet(list):
    """
    This class tries to implement the 'index'-fnc of the list class
    more efficiently. Using this, you can chance the value of item in a
    heap in an efficient manner.
    Only those functionalities used in pyheapq.py file  are implemented.
    If you have one item in this tructure more than once, this class
    does not work properly anymore..
    """
    def __init__(self, iterator = None):
        self.pos_dict = {}
        self.pop_begin = 0
        if iterator != None:
            for i in iterator:
                self.append(i)

    def append(self, item):
        if item in self.pos_dict:
            raise ValueError("item %s appended twice."%str(item))
        self.pos_dict[item] = self.pop_begin + len(self)
        list.append(self, item)

    def __setitem__(self,pos, item):
        oldvalue = self.__getitem__(pos)
        if self.pos_dict[oldvalue] == pos:
            del self.pos_dict[oldvalue]

        list.__setitem__(self,pos, item)
        self.pos_dict[item] = self.pop_begin + pos


    def pop(self, pos = None):
        if pos == None:
            item = list.pop(self)
            pos = len(self)
        elif pos == 0:
            item = list.pop(self,0)
            self.pop_begin += 1
        else:
            raise ValueError('pop is implemented just for begin and end')

        del self.pos_dict[item]

        return item

    def index(self,item):
        return self.pos_dict[item] - self.pop_begin

    def __str__(self):
        return list.__str__(self) + str(self.pos_dict) + "pops %d"%self.pop_begin



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module defining wait related actions
"""
from pyherc.aspects import log_debug, log_info


class WaitAction():
    """
    Action for waiting

    .. versionadded:: 0.10
    """
    @log_debug
    def __init__(self, character, time_to_wait):
        """
        Default constructor

        :param character: character waiting
        :type character: Character
        :param time_to_wait: amount of ticks to wait
        :type time_to_wait: int
        """
        super().__init__()
        self.character = character
        self.time_to_wait = time_to_wait

    @log_info
    def execute(self):
        """
        Executes this action
        """
        self.character.tick = self.character.tick + self.time_to_wait

    @log_debug
    def is_legal(self):
        """
        Check if the action is possible to perform

        :returns: True if waiting is possible, false otherwise
        :rtype: Boolean
        """
        return True



# -*- coding: utf-8 -*-

#   Copyright 2010-2014 Tuukka Turto
#
#   This file is part of pyherc.
#
#   pyherc is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   pyherc is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with pyherc.  If not, see <http://www.gnu.org/licenses/>.

"""
Module for main map related functionality
"""
from random import Random

from herculeum.ui.controllers import MoveController
from herculeum.ui.gui.animations import AnimationFactory
from herculeum.ui.gui.eventdisplay import EventMessageWidget
from herculeum.ui.gui.widgets import (EffectsWidget, HitPointsWidget,
                                      SpellSelectorWidget, TimerAdapter)
from pyherc.data.model import DIED_IN_DUNGEON
from pyherc.rules import attack, cast, is_move_legal, move, pick_up, wait
from PyQt4.QtCore import (pyqtProperty, pyqtSignal, QAbstractAnimation,
                          QEasingCurve, QEvent, QObject, QPropertyAnimation,
                          QSequentialAnimationGroup, QSize, Qt, QTimer)
from PyQt4.QtGui import (QColor, QFont, QGraphicsPixmapItem, QGraphicsScene,
                         QGraphicsSimpleTextItem, QGraphicsView, QHBoxLayout,
                         QTransform, QVBoxLayout, QWidget)


class PlayMapWindow(QWidget):
    """
    Window for displaying playing world

    .. versionadded:: 0.5
    """
    def __init__(self, parent, model, surface_manager, action_factory, rng,
                 rules_engine, configuration):
        """
        Default constructor
        """
        super().__init__(parent)

        self.model = model
        self.surface_manager = surface_manager
        self.action_factory = action_factory
        self.rng = rng
        self.current_level = None
        self.configuration = configuration

        self.hit_points_widget = None
        self.message_widget = None
        self.map_widget = None
        self.effects_widget = None
        self.spell_selector = None

        self.__set_layout(model, surface_manager, action_factory, rng,
                          rules_engine, configuration)

    MenuRequested = pyqtSignal(name='MenuRequested')
    EndScreenRequested = pyqtSignal(name='EndScreenRequested')

    def __set_layout(self, model, surface_manager, action_factory, rng,
                     rules_engine, configuration):
        """
        Set layout of this window
        """
        layout = QVBoxLayout()
        status_layout = QHBoxLayout()

        self.hit_points_widget = HitPointsWidget(parent = self)
        self.effects_widget = EffectsWidget(parent = self,
                                            surface_manager = surface_manager)

        self.spell_selector = SpellSelectorWidget(parent = self,
                                                  surface_manager = surface_manager)

        self.map_widget = PlayMapWidget(parent = self,
                                        model = model,
                                        surface_manager = surface_manager,
                                        action_factory = action_factory,
                                        rng = rng,
                                        rules_engine = rules_engine,
                                        configuration = configuration)
        self.map_widget.MenuRequested.connect(self.on_menu_requested)
        self.map_widget.EndScreenRequested.connect(self.on_end_screen_requested)
        self.map_widget.NextSpellRequested.connect(self.on_next_spell)
        self.map_widget.PreviousSpellRequested.connect(self.on_previous_spell)

        self.message_widget = EventMessageWidget(parent = self)
        self.message_widget.setMaximumHeight(100)

        status_layout.addWidget(self.hit_points_widget)
        status_layout.addWidget(self.spell_selector)
        status_layout.addWidget(self.effects_widget)
        status_layout.addStretch()

        layout.addLayout(status_layout)
        layout.addWidget(self.map_widget)
        layout.addWidget(self.message_widget)
        self.setLayout(layout)
        self.resize(QSize(640, 480))

    def construct_scene(self):
        """
        Create scene to display
        """
        self.map_widget.construct_scene()
        self.model.player.register_event_listener(self.message_widget)
        self.message_widget.set_point_of_view(self.model.player)
        self.model.player.register_for_updates(self.effects_widget)

    def on_menu_requested(self):
        """
        Handle requesting menu window
        """
        self.MenuRequested.emit()

    def on_end_screen_requested(self):
        """
        Handle requesting end screen

        .. versionadded:: 0.8
        """
        self.EndScreenRequested.emit()

    def on_next_spell(self):
        """
        Handle selecting next spell

        .. versionadded:: 0.10
        """
        self.spell_selector.next_spell()

    def on_previous_spell(self):
        """
        Handle selecting previous spell

        .. versionadded:: 0.10
        """
        self.spell_selector.previous_spell()

zorder_floor = 0
zorder_item = 2
zorder_character = 5
zorder_wall = 10
zorder_ornament = 20

zorder_counter = 30

class PlayMapWidget(QWidget):
    """
    Widget for displaying playing world

    .. versionadded:: 0.5
    """
    def __init__(self, parent, model, surface_manager, action_factory, rng,
                 rules_engine, configuration):
        """
        Default constructor
        """
        super().__init__(parent)

        self.model = model
        self.scene = None
        self.surface_manager = surface_manager
        self.action_factory = action_factory
        self.rng = rng
        self.rules_engine = rules_engine
        self.configuration = configuration

        self.current_level = None
        self.view = None
        self.animation_adapters = []
        self.animation_timers = []

        for adapter in range(10):
            self.animation_adapters.append(TimerAdapter())
            self.animation_timers.append(QTimer(self))
            self.animation_timers[adapter].timeout.connect(self.animation_adapters[adapter].trigger_animations)
            self.animation_timers[adapter].start(450 + adapter * 10)

        self.animations = []
        self.move_controller = MoveController(action_factory = action_factory,
                                              rng = rng)

        self.__set_layout()
        self.keymap, self.move_key_map = self._construct_keymaps(
                                                        configuration.controls)

        self.animation_factory = AnimationFactory()

    MenuRequested = pyqtSignal(name='MenuRequested')
    EndScreenRequested = pyqtSignal(name='EndScreenRequested')
    NextSpellRequested = pyqtSignal(name='NextSpellRequested')
    PreviousSpellRequested = pyqtSignal(name='PreviousSpellRequested')


    def _construct_keymaps(self, config):
        """
        Construct keymaps for handling input
        """
        keymap = {}
        move_keymap = {}
        for key in config.move_left:
            keymap[key] = self._move
            move_keymap[key] = 7
        for key in config.move_up:
            keymap[key] = self._move
            move_keymap[key] = 1
        for key in config.move_right:
            keymap[key] = self._move
            move_keymap[key] = 3
        for key in config.move_down:
            keymap[key] = self._move
            move_keymap[key] = 5
        for key in config.start:
            keymap[key] = self._menu
        for key in config.action_a:
            keymap[key] = self._action_a
        for key in config.back:
            keymap[key] = self._back
        for key in config.left_shoulder:
            keymap[key] = self._shoulder_left
        for key in config.right_shoulder:
            keymap[key] = self._shoulder_right

        return keymap, move_keymap

    def __set_layout(self):
        """
        Set layout of this widget
        """
        self.scene = QGraphicsScene()

        layout = QHBoxLayout()

        self.view = QGraphicsView(self.scene)
        self.view.setFocusPolicy(Qt.StrongFocus)
        self.view.installEventFilter(self)
        layout.addWidget(self.view)

        self.setLayout(layout)

    def construct_scene(self):
        """
        Construct scene to display
        """
        self.__construct_scene(self.model, self.scene)

        self.model.player.register_for_updates(self)
        self.model.register_event_listener(self)
        self.__center_view_on_character(self.model.player)

    def __center_view_on_character(self, entity):
        """
        Center view on given entity
        """
        location = entity.location
        width = 320
        height = 320

        self.view.setSceneRect((location[0] * 32) - width // 2,
                              (location[1] * 32) - height // 2,
                              width,
                              height)

    def __construct_scene(self, model, scene):
        """
        Constructs scene to display
        """
        for item in (item for item in scene.items()
                     if hasattr(item, 'clear_update_registration')):
            item.clear_update_registration()

        for adapter in self.animation_adapters:
            adapter.glyphs.clear()

        for anim in [x for x in self.animations]:
            anim.clear()

        self.animations = []

        scene.clear()

        self.current_level = model.player.level

        for loc_x, column in enumerate(self.current_level.floor):
            for loc_y, tile in enumerate(column):
                new_glyph = MapGlyph(self.surface_manager.get_icon(tile),
                                     None)
                new_glyph.setZValue(zorder_floor)
                new_glyph.setPos(loc_x * 32, loc_y * 32)
                scene.addItem(new_glyph)

        for loc_x, column in enumerate(self.current_level.walls):
            for loc_y, tile in enumerate(column):
                new_glyph = MapGlyph(self.surface_manager.get_icon(tile),
                                     None)
                new_glyph.setZValue(zorder_wall)
                new_glyph.setPos(loc_x * 32, loc_y * 32)
                scene.addItem(new_glyph)

        for loc_x, column in enumerate(self.current_level.ornamentations):
            for loc_y, tile in enumerate(column):
                if tile:
                    new_glyph = MapGlyph(self.surface_manager.get_icon(tile),
                                         None,
                                         self.rng.choice(self.animation_adapters))
                    new_glyph.setZValue(zorder_ornament)
                    new_glyph.setPos(loc_x * 32, loc_y * 32)
                    scene.addItem(new_glyph)

        for item in self.current_level.items:
            self.add_glyph(item, scene, zorder_item)

        for creature in self.current_level.creatures:
            self.add_glyph(creature,
                           scene,
                           zorder_character,
                           self.rng.choice(self.animation_adapters))

    def add_glyph(self, entity, scene, z_order, animation_adapter = None):
        """
        Add graphical representation of an entity

        :param entity: entity to display
        :param scene: scene where glyph will be added
        :type scene: QGraphicsScene
        :param z_order: z-order of entity being displayed
        :type z_order: int
        :param animation_adapter: adapter to register for animations
        :type animation_adapter: TimerAdapter
        """
        new_glyph = MapGlyph(self.surface_manager.get_icon(entity.icon),
                             entity, animation_adapter)
        new_glyph.setZValue(z_order)
        new_glyph.setPos(entity.location[0] * 32,
                         entity.location[1] * 32)
        scene.addItem(new_glyph)

    def receive_event(self, event):
        """
        Receive event from model
        """
        anim = self.animation_factory.create_animation(event)
        anim.trigger(self)

    def remove_finished_animation(self):
        """
        Remove finished animation
        """
        finished_animations  = (x for x in self.animations
                                if x.state() == QAbstractAnimation.Stopped)
        counters = (x.animationAt(0).targetObject().object_to_animate
                    for x in finished_animations)

        for item in finished_animations:
            item.clear()
            self.animations.remove(item)

        for item in counters:
            self.view.scene().removeItem(item)

    def receive_update(self, event):
        """
        Receive update from entity
        """
        if event.event_type == 'move':
            if self.model.player.level != self.current_level:
                self.__construct_scene(self.model, self.scene)
                self.__center_view_on_character(self.model.player)

    def eventFilter(self, qobject, event): #pylint: disable-msg=C0103
        """
        Filter events

        .. Note:: This is done in order to process cursor keys
        """
        result = False

        if event.type() == QEvent.KeyPress:
            self.keyPressEvent(event)
            result = True
        else:
            result = super().eventFilter(qobject, event)

        return result

    def keyPressEvent(self, event): #pylint: disable-msg=C0103
        """
        Handle key events
        """
        if self.model.player is None:
            return

        key_code = event.key()

        player = self.model.player
        next_creature = self.model.get_next_creature(self.rules_engine)

        if next_creature == player:

            if key_code in self.keymap:
                self.keymap[key_code](key_code, event.modifiers())

        next_creature = self.model.get_next_creature(self.rules_engine)

        if next_creature is None:
            self.model.end_condition = DIED_IN_DUNGEON

        while (next_creature != player
                and next_creature != None
                and self.model.end_condition == 0):
            next_creature.act(model = self.model,
                              action_factory = self.action_factory,
                              rng = self.rng)
            next_creature = self.model.get_next_creature(self.rules_engine)

            if next_creature is None:
                self.model.end_condition = DIED_IN_DUNGEON

        if self.model.end_condition != 0:
            self.EndScreenRequested.emit()

    def _move(self, key, modifiers):
        """
        Process movement key

        :param key: key triggering the processing
        :type key: int
        """
        player = self.model.player
        direction = self.move_key_map[key]

        if modifiers & Qt.ControlModifier:
            if direction != 9:
                attack(player,
                       direction,
                       self.action_factory,
                       self.rng)
        elif modifiers & Qt.AltModifier:
            if direction != 9:
                cast(player,
                     direction,
                     'fireball',
                     self.action_factory)

        else:
            self.move_controller.move_or_attack(player,
                                                direction,
                                                'walk')

    def _menu(self, key, modifiers):
        """
        Process menu key

        :param key: key triggering the processing
        :type key: int
        """
        self.MenuRequested.emit()

    def _back(self, key, modifiers):
        """
        Process back key

        :param key: key triggering the processing
        :type key: int
        """
        wait(self.model.player,
             self.action_factory)

    def _shoulder_right(self, key, modifiers):
        """
        Process right shoulder button

        :param key: key triggering the processing
        :type key: int

        .. versionadded:: 0.10
        """
        self.NextSpellRequested.emit()

    def _shoulder_left(self, key, modifiers):
        """
        Process left shoulder button

        :param key: key triggering the processing
        :type key: int

        .. versionadded:: 0.10
        """
        self.PreviousSpellRequested.emit()

    def _action_a(self, key, modifiers):
        """
        Process action a key

        :param key: key triggering the processing
        :type key: int
        """
        player = self.model.player
        level = player.level
        items = level.get_items_at(player.location)

        if items != None and len(items) > 0:
            pick_up(player,
                    items[0],
                    self.action_factory)

        elif is_move_legal(player, 9, 'walk', self.action_factory):
            move(player, 9, self.action_factory)

class DamageCounter(QGraphicsSimpleTextItem):
    """
    Counter for showing damage

    .. versionadded:: 0.6
    """
    def __init__(self, damage, colour, parent):
        """
        Default constructor
        """
        super().__init__()

        font = QFont('Helvetica',
                     12,
                     QFont.Bold,
                     False)

        self.setText(str(damage))
        self.setBrush(QColor(colour))
        self.setFont(font)

        self.adapter = DamageCounterAdapter(self, self)

class DamageCounterAdapter(QObject):
    """
    Adapter for damage counter

    .. versionadded:: 0.6
    """
    def __init__(self, parent, object_to_animate):
        """
        Default constructor
        """
        super().__init__()
        self.object_to_animate = object_to_animate

    def __get_y_location(self):
        return self.object_to_animate.y()

    def __set_y_location(self, y):
        self.object_to_animate.setY(y)

    def __get_opacity(self):
        return self.object_to_animate.opacity()

    def __set_opacity(self, opacity):
        self.object_to_animate.setOpacity(opacity)

    y_location = pyqtProperty(int, __get_y_location, __set_y_location)
    opacity = pyqtProperty(float, __get_opacity, __set_opacity)

class MapGlyph(QGraphicsPixmapItem):
    """
    Widget to represent a glyph on map

    .. versionadded:: 0.5
    """
    def __init__(self, pixmap, entity, timer = None):
        """
        Default constructor
        """
        self.tiles = []

        if hasattr(pixmap, 'alphaChannel'):
            super().__init__(pixmap, None)
        else:
            super().__init__(pixmap[0], None)
            if timer:
                timer.register(self)
            self.tiles = pixmap

        self.entity = entity

        self.flipped = False
        self.offset = 0

    def animate(self, frame):
        """
        Move animation to given frame

        .. versionadded:: 0.10
        """
        self.setPixmap(self.tiles[frame])

    def receive_update(self, event):
        """
        Receive event from model
        """
        if event.event_type == 'move':
            location = event.mover.location
            direction = event.direction

            if direction in (2, 3, 4):
                self.flipped = True
                self.offset = 32

            if direction in (6, 7, 8):
                self.flipped = False
                self.offset = 0

            if location != None:
                self.setPos(location[0] * 32 + self.offset,
                            location[1] * 32)

            if self.flipped:
                self.setTransform(QTransform.fromScale(-1, 1))
            else:
                self.setTransform(QTransform.fromScale(1, 1))

    def clear_update_registration(self):
        """
        Clear update registrations
        """
        if self.entity != None:
            self.entity.remove_from_updates(self)
